"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import re
import logging
import json
import boto3

from botocore.exceptions import NoCredentialsError
from django.db import models, transaction
from django.conf import settings
from django.utils.translation import gettext_lazy as _
from django.dispatch import receiver
from django.db.models.signals import post_save

from io_storages.base_models import (
    ImportStorage,
    ImportStorageLink,
    ExportStorage,
    ExportStorageLink,
)
from io_storages.utils import get_uri_via_regex
from io_storages.s3.utils import get_client_and_resource, resolve_s3_url
from tasks.validation import ValidationError as TaskValidationError
from tasks.models import Annotation

logger = logging.getLogger(__name__)
logging.getLogger("botocore").setLevel(logging.CRITICAL)
boto3.set_stream_logger(level=logging.INFO)
url_scheme = "s3"


class S3StorageMixin(models.Model):
    bucket = models.TextField(_("bucket"), null=True, blank=True, help_text="S3 bucket name")
    prefix = models.TextField(_("prefix"), null=True, blank=True, help_text="S3 bucket prefix")
    regex_filter = models.TextField(
        _("regex_filter"),
        null=True,
        blank=True,
        help_text="Cloud storage regex for filtering objects",
    )
    use_blob_urls = models.BooleanField(
        _("use_blob_urls"), default=False, help_text="Interpret objects as BLOBs and generate URLs"
    )
    aws_access_key_id = models.TextField(
        _("aws_access_key_id"), null=True, blank=True, help_text="AWS_ACCESS_KEY_ID"
    )
    aws_secret_access_key = models.TextField(
        _("aws_secret_access_key"), null=True, blank=True, help_text="AWS_SECRET_ACCESS_KEY"
    )
    aws_session_token = models.TextField(
        _("aws_session_token"), null=True, blank=True, help_text="AWS_SESSION_TOKEN"
    )
    region_name = models.TextField(_("region_name"), null=True, blank=True, help_text="AWS Region")
    s3_endpoint = models.TextField(_("s3_endpoint"), null=True, blank=True, help_text="S3 Endpoint")

    def get_client_and_resource(self):
        return get_client_and_resource(
            self.aws_access_key_id,
            self.aws_secret_access_key,
            self.aws_session_token,
            self.region_name,
            self.s3_endpoint,
        )

    def get_client(self):
        client, _ = self.get_client_and_resource()
        return client

    def get_client_and_bucket(self, validate_connection=True):
        client, s3 = self.get_client_and_resource()
        if validate_connection:
            self.validate_connection(client)
        return client, s3.Bucket(self.bucket)

    def validate_connection(self, client=None):
        if client is None:
            client = self.get_client()
        if self.prefix:
            logger.debug(f"Test connection to bucket {self.bucket} with prefix {self.prefix}")
            result = client.list_objects_v2(Bucket=self.bucket, Prefix=self.prefix, MaxKeys=1)
            if not result.get("KeyCount"):
                raise KeyError(f"s3://{self.bucket}/{self.prefix} not found.")
        else:
            logger.debug(f"Test connection to bucket {self.bucket}")
            client.head_bucket(Bucket=self.bucket)

    @property
    def path_full(self):
        prefix = self.prefix or ""
        return f"{url_scheme}://{self.bucket}/{prefix}"

    @property
    def type_full(self):
        return "Amazon AWS S3"

    class Meta:
        abstract = True


class S3ImportStorage(S3StorageMixin, ImportStorage):

    presign = models.BooleanField(_("presign"), default=True, help_text="Generate presigned URLs")
    presign_ttl = models.PositiveSmallIntegerField(
        _("presign_ttl"), default=1, help_text="Presigned URLs TTL (in minutes)"
    )

    def iterkeys(self):
        client, bucket = self.get_client_and_bucket()
        if self.prefix:
            bucket_iter = bucket.objects.filter(
                Prefix=self.prefix.rstrip("/") + "/", Delimiter="/"
            ).all()
        else:
            bucket_iter = bucket.objects.all()
        regex = re.compile(str(self.regex_filter)) if self.regex_filter else None
        for obj in bucket_iter:
            key = obj.key
            if key.endswith("/"):
                logger.debug(key + " is skipped because it is a folder")
                continue
            if regex and not regex.match(key):
                logger.debug(key + " is skipped by regex filter")
                continue
            yield key

    def scan_and_create_links(self):
        return self._scan_and_create_links(S3ImportStorageLink)

    def _get_validated_task(self, parsed_data, key):
        """Validate parsed data with labeling config and task structure"""
        if not isinstance(parsed_data, dict):
            raise TaskValidationError(
                "Error at " + str(key) + ":\n"
                "Cloud storage supports one task (one dict object) per JSON file only. "
            )
        return parsed_data

    def get_data(self, key):
        uri = f"s3://{self.bucket}/{key}"
        if self.use_blob_urls:
            data_key = settings.DATA_UNDEFINED_NAME
            return {data_key: uri}

        # read task json from bucket and validate it
        _, s3 = self.get_client_and_resource()
        bucket = s3.Bucket(self.bucket)
        obj = s3.Object(bucket.name, key).get()["Body"].read().decode("utf-8")
        value = json.loads(obj)
        if not isinstance(value, dict):
            raise ValueError(
                f"Error on key {key}: For S3 your JSON file must be a dictionary with one task"
            )

        value = self._get_validated_task(value, key)
        return value

    def resolve_uri(self, data):
        try:
            uri, storage = get_uri_via_regex(data, prefixes=(url_scheme,))
            if not storage:
                return
            resolved_uri = resolve_s3_url(uri, self.get_client(), self.presign)
            return data.replace(uri, resolved_uri)
        except NoCredentialsError:
            logger.warning(f"No AWS credentials specified for {data}")


class S3ExportStorage(S3StorageMixin, ExportStorage):
    def save_annotation(self, annotation):
        client, s3 = self.get_client_and_resource()
        logger.debug(
            f"Creating new object on {self.__class__.__name__} Storage {self} for annotation {annotation}"
        )
        ser_annotation = self._get_serialized_data(annotation)
        with transaction.atomic():
            # Create export storage link
            link = S3ExportStorageLink.create(annotation, self)
            key = str(self.prefix) + "/" + link.key if self.prefix else link.key
            try:
                s3.Object(self.bucket, key).put(Body=json.dumps(ser_annotation))
            except Exception as exc:
                logger.error(
                    f"Can't export annotation {annotation} to S3 storage {self}. Reason: {exc}",
                    exc_info=True,
                )


@receiver(post_save, sender=Annotation)
def export_annotation_to_s3_storages(sender, instance, **kwargs):
    project = instance.task.project
    if hasattr(project, "io_storages_s3exportstorages"):
        for storage in project.io_storages_s3exportstorages.all():
            logger.debug(f"Export {instance} to S3 storage {storage}")
            storage.save_annotation(instance)


class S3ImportStorageLink(ImportStorageLink):
    storage = models.ForeignKey(S3ImportStorage, on_delete=models.CASCADE, related_name="links")


class S3ExportStorageLink(ExportStorageLink):
    storage = models.ForeignKey(S3ExportStorage, on_delete=models.CASCADE, related_name="links")



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import ujson as json

from rest_framework import serializers
from drf_dynamic_fields import DynamicFieldsMixin

from organizations.models import Organization, OrganizationMember
from users.serializers import UserSerializer
from collections import OrderedDict


class OrganizationIdSerializer(DynamicFieldsMixin, serializers.ModelSerializer):
    class Meta:
        model = Organization
        fields = ["id", "title"]


class OrganizationSerializer(DynamicFieldsMixin, serializers.ModelSerializer):
    class Meta:
        model = Organization
        fields = "__all__"


class OrganizationMemberSerializer(DynamicFieldsMixin, serializers.ModelSerializer):
    class Meta:
        model = OrganizationMember
        fields = ["id", "organization", "user"]


class UserSerializerWithProjects(UserSerializer):
    created_projects = serializers.SerializerMethodField(read_only=True)
    contributed_to_projects = serializers.SerializerMethodField(read_only=True)

    def get_created_projects(self, user):
        return user.created_projects.order_by("created_at").values("id", "title")

    def get_contributed_to_projects(self, user):
        projects = user.annotations.order_by("created_at").values(
            "task__project__id", "task__project__title"
        )
        contributed_to = [
            (json.dumps({"id": p["task__project__id"], "title": p["task__project__title"]}), 0)
            for p in projects
        ]
        contributed_to = OrderedDict(contributed_to)  # remove duplicates without ordering losing
        return [json.loads(key) for key in contributed_to]

    class Meta(UserSerializer.Meta):
        fields = UserSerializer.Meta.fields + ("created_projects", "contributed_to_projects")


class OrganizationMemberUserSerializer(DynamicFieldsMixin, serializers.ModelSerializer):
    """Adds all user properties"""

    user = UserSerializerWithProjects()

    class Meta:
        model = OrganizationMember
        fields = ["id", "organization", "user"]


class OrganizationInviteSerializer(serializers.Serializer):
    token = serializers.CharField(required=False)
    invite_url = serializers.CharField(required=False)



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
from django.urls import include, path

from organizations import api, views

app_name = "organizations"

# TODO: there should be only one patterns list based on API (with api/ prefix removed)
# Page URLs
_urlpatterns = [
    # get organization page
    path("", views.organization_people_list, name="organization-index"),
]

# API URLs
_api_urlpattens = [
    # organization list viewset
    path("", api.OrganizationListAPI.as_view(), name="organization-index"),
    # organization detail viewset
    path("<int:pk>", api.OrganizationAPI.as_view(), name="organization-detail"),
    # organization memberships list viewset
    path(
        "<int:pk>/memberships",
        api.OrganizationMemberListAPI.as_view(),
        name="organization-memberships-list",
    ),
]

# TODO: these urlpatterns should be moved in core/urls with include('organizations.urls')
urlpatterns = [
    path("people/", include(_urlpatterns)),
    path("api/organizations/", include((_api_urlpattens, app_name), namespace="api")),
    # invite
    path("api/invite", api.OrganizationInviteAPI.as_view(), name="organization-reset-token"),
    path(
        "api/invite/reset-token",
        api.OrganizationResetTokenAPI.as_view(),
        name="organization-reset-token",
    ),
]



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
from django.urls import path, include

from . import api

app_name = "ml"

# ML backend CRUD
_api_urlpatterns = [
    # All storages
    path("", api.MLBackendListAPI.as_view(), name="ml-list"),
    path("<int:pk>", api.MLBackendDetailAPI.as_view(), name="ml-detail"),
    path("<int:pk>/train", api.MLBackendTrainAPI.as_view(), name="ml-train"),
]

urlpatterns = [
    path("api/ml/", include((_api_urlpatterns, app_name), namespace="api")),
]



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import os
import io
import csv
import ssl
import uuid
import pickle
import logging

try:
    import ujson as json
except:
    import json
import pandas as pd
import numpy as np

from dateutil import parser
from rest_framework.exceptions import ValidationError
from django.conf import settings
from django.core.files.uploadedfile import SimpleUploadedFile
from urllib.request import urlopen

from .models import FileUpload

logger = logging.getLogger(__name__)
csv.field_size_limit(131072 * 10)


def is_binary(f):
    return isinstance(f, (io.RawIOBase, io.BufferedIOBase))


def csv_generate_header(file):
    """Generate column names for headless csv file"""
    file.seek(0)
    names = []
    line = file.readline()

    num_columns = len(line.split(b"," if isinstance(line, bytes) else ","))
    for i in range(num_columns):
        names.append("column" + str(i + 1))
    file.seek(0)
    return names


def check_max_task_number(tasks):
    # max tasks
    if len(tasks) > settings.TASKS_MAX_NUMBER:
        raise ValidationError(
            f"Maximum task number is {settings.TASKS_MAX_NUMBER}, "
            f"current task number is {len(tasks)}"
        )


def check_file_sizes_and_number(files):
    total = sum([file.size for _, file in files.items()])

    if total >= settings.TASKS_MAX_FILE_SIZE:
        raise ValidationError(
            f"Maximum total size of all files is {settings.TASKS_MAX_FILE_SIZE} bytes, "
            f"current size is {total} bytes"
        )


def create_file_upload(request, project, file):
    instance = FileUpload(user=request.user, project=project, file=file)
    instance.save()
    return instance


def load_tasks(request, project):
    """Load tasks from different types of request.data / request.files"""
    file_upload_ids, found_formats, data_keys = [], [], set()
    could_be_tasks_lists = False
    # take tasks from request FILES
    if len(request.FILES):
        check_file_sizes_and_number(request.FILES)
        for filename, file in request.FILES.items():
            file_upload = create_file_upload(request, project, file)
            if file_upload.format_could_be_tasks_list:
                could_be_tasks_lists = True
            file_upload_ids.append(file_upload.id)
        tasks, found_formats, data_keys = FileUpload.load_tasks_from_uploaded_files(
            project, file_upload_ids
        )

    # take tasks from url address
    elif "application/x-www-form-urlencoded" in request.content_type:
        ctx = ssl.create_default_context()
        ctx.check_hostname = False
        ctx.verify_mode = ssl.CERT_NONE

        try:
            url = request.data["url"]
            filename = url.rsplit("/", 1)[-1]
            with urlopen(url, context=ctx) as file:
                # check size
                meta = file.info()
                file.size = int(meta.get("Content-Length"))
                file.urlopen = True
                check_file_sizes_and_number({url: file})
                file_content = file.read()
                if isinstance(file_content, str):
                    file_content = file_content.encode()
                file_upload = create_file_upload(
                    request, project, SimpleUploadedFile(filename, file_content)
                )
                file_upload_ids.append(file_upload.id)
                tasks, found_formats, data_keys = FileUpload.load_tasks_from_uploaded_files(
                    project, file_upload_ids
                )

        except ValidationError as e:
            raise e
        except Exception as e:
            raise ValidationError(str(e))

    # take one task from request DATA
    elif "application/json" in request.content_type and isinstance(request.data, dict):
        tasks = [request.data]

        # take many tasks from request DATA
    elif "application/json" in request.content_type and isinstance(request.data, list):
        tasks = request.data

    # incorrect data source
    else:
        raise ValidationError("load_tasks: No data found in DATA or in FILES")

    # check is data root is list
    if not isinstance(tasks, list):
        raise ValidationError("load_tasks: Data root must be list")

    # empty tasks error
    if not tasks:
        raise ValidationError("load_tasks: No tasks added")

    check_max_task_number(tasks)
    return tasks, file_upload_ids, could_be_tasks_lists, found_formats, list(data_keys)



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
from django.utils.decorators import method_decorator
from drf_yasg.utils import swagger_auto_schema

from io_storages.s3.models import S3ImportStorage, S3ExportStorage
from io_storages.s3.serializers import S3ImportStorageSerializer, S3ExportStorageSerializer
from io_storages.api import (
    ImportStorageListAPI,
    ImportStorageDetailAPI,
    ImportStorageSyncAPI,
    ExportStorageListAPI,
    ExportStorageDetailAPI,
    ImportStorageValidateAPI,
    ExportStorageValidateAPI,
    ImportStorageFormLayoutAPI,
    ExportStorageFormLayoutAPI,
)


@method_decorator(
    name="get",
    decorator=swagger_auto_schema(
        tags=["Storage"],
        operation_summary="S3: Get import storage",
        operation_description="Get list of all S3 import storage connections.",
    ),
)
@method_decorator(
    name="post",
    decorator=swagger_auto_schema(
        tags=["Storage"],
        operation_summary="S3: Create new storage",
        operation_description="Get new S3 import storage",
    ),
)
class S3ImportStorageListAPI(ImportStorageListAPI):
    queryset = S3ImportStorage.objects.all()
    serializer_class = S3ImportStorageSerializer


@method_decorator(
    name="get",
    decorator=swagger_auto_schema(
        tags=["Storage"],
        operation_summary="S3: Get import storage",
        operation_description="Get a specific S3 import storage connection.",
    ),
)
@method_decorator(
    name="patch",
    decorator=swagger_auto_schema(
        tags=["Storage"],
        operation_summary="S3: Update import storage",
        operation_description="Update a specific S3 import storage connection.",
    ),
)
@method_decorator(
    name="delete",
    decorator=swagger_auto_schema(
        tags=["Storage"],
        operation_summary="S3: Delete import storage",
        operation_description="Delete a specific S3 import storage connection.",
    ),
)
class S3ImportStorageDetailAPI(ImportStorageDetailAPI):
    queryset = S3ImportStorage.objects.all()
    serializer_class = S3ImportStorageSerializer


@method_decorator(
    name="post",
    decorator=swagger_auto_schema(
        tags=["Storage"],
        operation_summary="S3: Sync import storage",
        operation_description="Sync tasks from an S3 import storage connection.",
    ),
)
class S3ImportStorageSyncAPI(ImportStorageSyncAPI):
    serializer_class = S3ImportStorageSerializer


@method_decorator(
    name="post",
    decorator=swagger_auto_schema(
        tags=["Storage"],
        operation_summary="S3: Validate import storage",
        operation_description="Validate a specific S3 import storage connection.",
    ),
)
class S3ImportStorageValidateAPI(ImportStorageValidateAPI):
    serializer_class = S3ImportStorageSerializer


@method_decorator(
    name="post",
    decorator=swagger_auto_schema(
        tags=["Storage"],
        operation_summary="S3: Validate export storage",
        operation_description="Validate a specific S3 export storage connection.",
    ),
)
class S3ExportStorageValidateAPI(ExportStorageValidateAPI):
    serializer_class = S3ExportStorageSerializer


@method_decorator(
    name="get",
    decorator=swagger_auto_schema(
        tags=["Storage"],
        operation_summary="S3: Get all export storage",
        operation_description="Get a list of all S3 export storage connections.",
    ),
)
@method_decorator(
    name="post",
    decorator=swagger_auto_schema(
        tags=["Storage"],
        operation_summary="S3: Create export storage",
        operation_description="Create a new S3 export storage connection to store annotations.",
    ),
)
class S3ExportStorageListAPI(ExportStorageListAPI):
    queryset = S3ExportStorage.objects.all()
    serializer_class = S3ExportStorageSerializer


@method_decorator(
    name="get",
    decorator=swagger_auto_schema(
        tags=["Storage"],
        operation_summary="S3: Get export storage",
        operation_description="Get a specific S3 export storage connection.",
    ),
)
@method_decorator(
    name="patch",
    decorator=swagger_auto_schema(
        tags=["Storage"],
        operation_summary="S3: Update export storage",
        operation_description="Update a specific S3 export storage connection.",
    ),
)
@method_decorator(
    name="delete",
    decorator=swagger_auto_schema(
        tags=["Storage"],
        operation_summary="S3: Delete export storage",
        operation_description="Delete a specific S3 export storage connection.",
    ),
)
class S3ExportStorageDetailAPI(ExportStorageDetailAPI):
    queryset = S3ExportStorage.objects.all()
    serializer_class = S3ExportStorageSerializer


class S3ImportStorageFormLayoutAPI(ImportStorageFormLayoutAPI):
    pass


class S3ExportStorageFormLayoutAPI(ExportStorageFormLayoutAPI):
    pass



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import os
import pkg_resources
import shutil
import glob
import io
import ujson as json
import yaml

from contextlib import contextmanager
from tempfile import mkstemp, mkdtemp

from appdirs import user_config_dir, user_data_dir, user_cache_dir


_DIR_APP_NAME = "label-studio"


def good_path(path):
    return os.path.abspath(os.path.expanduser(path))


def find_node(package_name, node_path, node_type):
    assert node_type in ("dir", "file", "any")
    basedir = pkg_resources.resource_filename(package_name, "")
    node_path = os.path.join(*node_path.split("/"))  # linux to windows compatibility
    search_by_path = "/" in node_path or "\\" in node_path

    for path, dirs, filenames in os.walk(basedir):
        if node_type == "file":
            nodes = filenames
        elif node_type == "dir":
            nodes = dirs
        else:
            nodes = filenames + dirs
        if search_by_path:
            for found_node in nodes:
                found_node = os.path.join(path, found_node)
                if found_node.endswith(node_path):
                    return found_node
        elif node_path in nodes:
            return os.path.join(path, node_path)
    else:
        raise IOError('Could not find "%s" at package "%s"' % (node_path, basedir))


def find_file(file):
    return find_node("label_studio_withoutsignin", file, "file")


def find_dir(directory):
    return find_node("label_studio_withoutsignin", directory, "dir")


@contextmanager
def get_temp_file():
    fd, path = mkstemp()
    yield path
    os.close(fd)


@contextmanager
def get_temp_dir():
    dirpath = mkdtemp()
    yield dirpath
    shutil.rmtree(dirpath)


def get_config_dir():
    config_dir = user_config_dir(appname=_DIR_APP_NAME)
    os.makedirs(config_dir, exist_ok=True)
    return config_dir


def get_data_dir():
    data_dir = user_data_dir(appname=_DIR_APP_NAME)
    os.makedirs(data_dir, exist_ok=True)
    return data_dir


def get_cache_dir():
    cache_dir = user_cache_dir(appname=_DIR_APP_NAME)
    os.makedirs(cache_dir, exist_ok=True)
    return cache_dir


def delete_dir_content(dirpath):
    for f in glob.glob(dirpath + "/*"):
        remove_file_or_dir(f)


def remove_file_or_dir(path):
    if os.path.isfile(path):
        os.remove(path)
    elif os.path.isdir(path):
        shutil.rmtree(path)


def get_all_files_from_dir(d):
    out = []
    for name in os.listdir(d):
        filepath = os.path.join(d, name)
        if os.path.isfile(filepath):
            out.append(filepath)
    return out


def iter_files(root_dir, ext):
    for root, _, files in os.walk(root_dir):
        for f in files:
            if f.lower().endswith(ext):
                yield os.path.join(root, f)


def json_load(file, int_keys=False):
    with io.open(file, encoding="utf8") as f:
        data = json.load(f)
        if int_keys:
            return {int(k): v for k, v in data.items()}
        else:
            return data


def read_yaml(filepath):
    if not os.path.exists(filepath):
        filepath = find_file(filepath)
    with io.open(filepath, encoding="utf-8") as f:
        data = yaml.load(f, Loader=yaml.FullLoader)
    return data


def read_bytes_stream(filepath):
    with open(filepath, mode="rb") as f:
        return io.BytesIO(f.read())



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import logging

from collections import OrderedDict
from django.conf import settings
from rest_framework.generics import get_object_or_404

from core.utils.common import int_from_request
from data_manager.prepare_params import PrepareParams
from data_manager.models import View
from tasks.models import Task


DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"
TASKS = "tasks:"
logger = logging.getLogger(__name__)


class DataManagerException(Exception):
    pass


def get_all_columns(project):
    """Make columns info for the frontend data manager"""
    result = {"columns": []}

    # frontend uses MST data model, so we need two directional referencing parent <-> child
    task_data_children = []
    i = 0

    data_types = OrderedDict()
    # add data types from config again
    data_types.update(project.data_types.items())
    # all data types from import data
    if project.summary.all_data_columns:
        data_types.update(
            {key: "Unknown" for key in project.summary.all_data_columns if key not in data_types}
        )

    # remove $undefined$ if there is one type at least in labeling config, because it will be resolved automatically
    if len(project.data_types) > 0:
        data_types.pop(settings.DATA_UNDEFINED_NAME, None)

    for key, data_type in list(data_types.items()):  # make data types from labeling config first
        column = {
            "id": key,
            "title": key if key != settings.DATA_UNDEFINED_NAME else "data",
            "type": data_type
            if data_type in ["Image", "Audio", "AudioPlus", "Unknown"]
            else "String",
            "target": "tasks",
            "parent": "data",
            "visibility_defaults": {
                "explore": True,
                "labeling": key in project.data_types or key == settings.DATA_UNDEFINED_NAME,
            },
        }
        result["columns"].append(column)
        task_data_children.append(column["id"])
        i += 1

    # --- Data root ---
    data_root = {
        "id": "data",
        "title": "data",
        "type": "List",
        "target": "tasks",
        "children": task_data_children,
    }

    result["columns"] += [
        # --- Tasks ---
        {
            "id": "id",
            "title": "ID",
            "type": "Number",
            "help": "Task ID",
            "target": "tasks",
            "visibility_defaults": {"explore": True, "labeling": False},
        },
        {
            "id": "completed_at",
            "title": "Completed",
            "type": "Datetime",
            "target": "tasks",
            "help": "Last annotation date",
            "visibility_defaults": {"explore": True, "labeling": False},
        },
        {
            "id": "total_annotations",
            "title": "Annotations",
            "type": "Number",
            "target": "tasks",
            "help": "Total annotations per task",
            "visibility_defaults": {"explore": True, "labeling": True},
        },
        {
            "id": "cancelled_annotations",
            "title": "Cancelled",
            "type": "Number",
            "target": "tasks",
            "help": "Total cancelled (skipped) annotations",
            "visibility_defaults": {"explore": True, "labeling": False},
        },
        {
            "id": "total_predictions",
            "title": "Predictions",
            "type": "Number",
            "target": "tasks",
            "help": "Total predictions per task",
            "visibility_defaults": {"explore": True, "labeling": False},
        },
        {
            "id": "annotations_results",
            "title": "Annotation results",
            "type": "String",
            "target": "tasks",
            "help": "Annotation results stacked over all annotations",
            "visibility_defaults": {"explore": False, "labeling": False},
        },
        {
            "id": "predictions_score",
            "title": "Prediction score",
            "type": "Number",
            "target": "tasks",
            "help": "Average prediction score over all task predictions",
            "visibility_defaults": {"explore": False, "labeling": False},
        },
        {
            "id": "predictions_results",
            "title": "Prediction results",
            "type": "String",
            "target": "tasks",
            "help": "Prediction results stacked over all predictions",
            "visibility_defaults": {"explore": False, "labeling": False},
        },
        {
            "id": "file_upload",
            "title": "Source filename",
            "type": "String",
            "target": "tasks",
            "help": "Source filename from import step",
            "visibility_defaults": {"explore": False, "labeling": False},
        },
        {
            "id": "created_at",
            "title": "Created at",
            "type": "Datetime",
            "target": "tasks",
            "help": "Task creation time",
            "visibility_defaults": {"explore": False, "labeling": False},
        },
        {
            "id": "annotators",
            "title": "Annotated by",
            "type": "List",
            "target": "tasks",
            "help": "All users who completed the task",
            "schema": {"items": project.organization.members.values_list("user__id", flat=True)},
            "visibility_defaults": {"explore": True, "labeling": False},
        },
    ]

    result["columns"].append(data_root)

    return result


def get_prepared_queryset(request, project):
    # use filters and selected items from view
    view_id = int_from_request(request.GET, "view_id", 0)
    if view_id > 0:
        view = get_object_or_404(request, View, pk=view_id)
        if view.project.pk != project.pk:
            raise DataManagerException("Project and View mismatch")
        prepare_params = view.get_prepare_tasks_params(add_selected_items=True)

    # use filters and selected items from request if it's specified
    else:
        selected = request.data.get("selectedItems", {"all": True, "excluded": []})
        if not isinstance(selected, dict):
            raise DataManagerException(
                'selectedItems must be dict: {"all": [true|false], '
                '"excluded | included": [...task_ids...]}'
            )
        filters = request.data.get("filters", None)
        ordering = request.data.get("ordering", [])
        prepare_params = PrepareParams(
            project=project.id,
            selectedItems=selected,
            data=request.data,
            filters=filters,
            ordering=ordering,
        )

    queryset = Task.prepared.all(prepare_params=prepare_params)
    return queryset


def evaluate_predictions(tasks):
    """Call ML backend for prediction evaluation of the task queryset"""
    if not tasks:
        return

    project = tasks[0].project

    for ml_backend in project.ml_backends.all():
        # tasks = tasks.filter(~Q(predictions__model_version=ml_backend.model_version))
        ml_backend.predict_many_tasks(tasks)



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import time
import logging
import drf_yasg.openapi as openapi
import json

from django.conf import settings
from django.db import transaction
from drf_yasg.utils import swagger_auto_schema
from rest_framework import generics, status
from rest_framework.response import Response
from rest_framework.parsers import FormParser, JSONParser, MultiPartParser
from rest_framework.permissions import IsAuthenticated
from ranged_fileresponse import RangedFileResponse

from core.permissions import all_permissions, ViewClassPermission
from core.utils.common import bool_from_request, retry_database_locked
from projects.models import Project
from tasks.models import Task
from .uploader import load_tasks
from .serializers import ImportApiSerializer, FileUploadSerializer
from .models import FileUpload

logger = logging.getLogger(__name__)


task_create_response_scheme = {
    201: openapi.Response(
        description="Tasks successfully imported",
        schema=openapi.Schema(
            title="Task creation response",
            description="Task creation response",
            type=openapi.TYPE_OBJECT,
            properties={
                "task_count": openapi.Schema(
                    title="task_count",
                    description="Number of tasks added",
                    type=openapi.TYPE_INTEGER,
                ),
                "annotation_count": openapi.Schema(
                    title="annotation_count",
                    description="Number of annotations added",
                    type=openapi.TYPE_INTEGER,
                ),
                "predictions_count": openapi.Schema(
                    title="predictions_count",
                    description="Number of predictions added",
                    type=openapi.TYPE_INTEGER,
                ),
                "duration": openapi.Schema(
                    title="duration",
                    description="Time in seconds to create",
                    type=openapi.TYPE_NUMBER,
                ),
                "file_upload_ids": openapi.Schema(
                    title="file_upload_ids",
                    description="Database IDs of uploaded files",
                    type=openapi.TYPE_ARRAY,
                    items=openapi.Schema(title="File Upload IDs", type=openapi.TYPE_INTEGER),
                ),
                "could_be_tasks_list": openapi.Schema(
                    title="could_be_tasks_list",
                    description="Whether uploaded files can contain lists of tasks, like CSV/TSV files",
                    type=openapi.TYPE_BOOLEAN,
                ),
                "found_formats": openapi.Schema(
                    title="found_formats",
                    description="The list of found file formats",
                    type=openapi.TYPE_ARRAY,
                    items=openapi.Schema(title="File format", type=openapi.TYPE_STRING),
                ),
                "data_columns": openapi.Schema(
                    title="data_columns",
                    description="The list of found data columns",
                    type=openapi.TYPE_ARRAY,
                    items=openapi.Schema(title="Data column name", type=openapi.TYPE_STRING),
                ),
            },
        ),
    ),
    400: openapi.Schema(
        title="Incorrect task data",
        description="String with error description",
        type=openapi.TYPE_STRING,
    ),
}


# Import
class ImportAPI(generics.CreateAPIView):
    """post:
    Import tasks

    Importing data as labeling tasks in bulk using this API endpoint works the same as using the Import button on the 
    Data Manager page. You can use this API endpoint for importing multiple tasks. One POST request is limited at 250K tasks and 200 MB.

    **Note:** Imported data is verified against a project *label_config* and must
    include all variables that were used in the *label_config*. For example,
    if the label configuration has a *$text* variable, then each item in a data object
    must include a "text" field.


    <br>

    ## POST requests

    <hr style="opacity:0.3">

    There are three possible ways to import tasks with this endpoint:

    1\. **POST with data**: Send JSON tasks as POST data. Only JSON is supported for POSTing files directly.

    ```bash
    curl -H 'Content-Type: application/json' -H 'Authorization: Token abc123' \\
    -X POST 'http://localhost/api/projects/1/import' --data '[{"text": "Some text 1"}, {"text": "Some text 2"}]'
    ```

     2\. **POST with files**: Send tasks as files. You can attach multiple files with different names. 


    - **JSON**: text files in javascript object notation format
    - **CSV**: text files with tables in Comma Separated Values format
    - **TSV**: text files with tables in Tab Separated Value format
    - **TXT**: simple text files are similar to CSV with one column and no header, supported for projects with one source only

    ```bash
    curl -H 'Content-Type: application/json' -H 'Authorization: Token abc123' \\
    -X POST 'http://localhost/api/projects/1/import' -F ‘file=@path/to/my_file.csv’
    ```

    3\. **POST with URL**: You can also provide a URL to a file with labeling tasks. Supported file formats are the same as in option 2.

    ```bash
    curl -H 'Content-Type: application/json' -H 'Authorization: Token abc123' \\
    -X POST 'http://localhost/api/projects/1/import' \\
    --data '[{"url": "http://example.com/test1.csv"}, {"url": "http://example.com/test2.csv"}]'
    ```

    <br>

    """

    permission_required = all_permissions.projects_change
    parser_classes = (JSONParser, MultiPartParser, FormParser)
    serializer_class = ImportApiSerializer
    queryset = Task.objects.all()

    def get_serializer_context(self):
        project_id = self.kwargs.get("pk")
        if project_id:
            project = generics.get_object_or_404(
                Project.objects.for_user(self.request.user), pk=project_id
            )
        else:
            project = None
        return {"project": project, "user": self.request.user}

    @swagger_auto_schema(tags=["Import"], responses=task_create_response_scheme)
    def post(self, *args, **kwargs):
        return super(ImportAPI, self).post(*args, **kwargs)

    def _save(self, tasks):
        serializer = self.get_serializer(data=tasks, many=True)
        serializer.is_valid(raise_exception=True)
        return serializer.save(project_id=self.kwargs["pk"]), serializer

    def create(self, request, *args, **kwargs):
        start = time.time()
        commit_to_project = bool_from_request(request.query_params, "commit_to_project", True)

        # check project permissions
        project = generics.get_object_or_404(
            Project.objects.for_user(self.request.user), pk=self.kwargs["pk"]
        )

        # upload files from request, and parse all tasks
        (
            parsed_data,
            file_upload_ids,
            could_be_tasks_lists,
            found_formats,
            data_columns,
        ) = load_tasks(request, project)

        if commit_to_project:
            # Immediately create project tasks and update project states and counters
            tasks, serializer = self._save(parsed_data)
            task_count = len(tasks)
            annotation_count = len(serializer.db_annotations)
            prediction_count = len(serializer.db_predictions)
            # Update tasks states if there are related settings in project
            # after bulk create we can bulk update tasks stats with
            # flag_update_stats=True but they are already updated with signal in same transaction
            # so just update tasks_number_changed
            project.update_tasks_states(
                maximum_annotations_changed=False,
                overlap_cohort_percentage_changed=False,
                tasks_number_changed=True,
            )
            logger.info("Tasks bulk_update finished")

            project.summary.update_data_columns(parsed_data)
            # TODO: project.summary.update_created_annotations_and_labels
        else:
            # Do nothing - just output file upload ids for further use
            task_count = len(parsed_data)
            annotation_count = None
            prediction_count = None

        duration = time.time() - start

        return Response(
            {
                "task_count": task_count,
                "annotation_count": annotation_count,
                "prediction_count": prediction_count,
                "duration": duration,
                "file_upload_ids": file_upload_ids,
                "could_be_tasks_list": could_be_tasks_lists,
                "found_formats": found_formats,
                "data_columns": data_columns,
            },
            status=status.HTTP_201_CREATED,
        )


class TasksBulkCreateAPI(ImportAPI):
    # just for compatibility - can be safely removed
    swagger_schema = None


class ReImportAPI(ImportAPI):
    """post:
    Re-import tasks

    Re-import tasks using the specified file upload IDs for a specific project.
    """

    permission_required = all_permissions.projects_change

    @retry_database_locked()
    def create(self, request, *args, **kwargs):
        start = time.time()
        files_as_tasks_list = bool_from_request(request.data, "files_as_tasks_list", True)

        # check project permissions
        project = generics.get_object_or_404(
            Project.objects.for_user(self.request.user), pk=self.kwargs["pk"]
        )
        file_upload_ids = self.request.data.get("file_upload_ids")
        tasks, found_formats, data_columns = FileUpload.load_tasks_from_uploaded_files(
            project, file_upload_ids, files_as_tasks_list=files_as_tasks_list
        )

        with transaction.atomic():
            project.remove_tasks_by_file_uploads(file_upload_ids)
            tasks, serializer = self._save(tasks)
        duration = time.time() - start

        # Update task states if there are related settings in project
        # after bulk create we can bulk update task stats with
        # flag_update_stats=True but they are already updated with signal in same transaction
        # so just update tasks_number_changed
        project.update_tasks_states(
            maximum_annotations_changed=False,
            overlap_cohort_percentage_changed=False,
            tasks_number_changed=True,
        )
        logger.info("Tasks bulk_update finished")

        project.summary.update_data_columns(tasks)
        # TODO: project.summary.update_created_annotations_and_labels

        return Response(
            {
                "task_count": len(tasks),
                "annotation_count": len(serializer.db_annotations),
                "prediction_count": len(serializer.db_predictions),
                "duration": duration,
                "file_upload_ids": file_upload_ids,
                "found_formats": found_formats,
                "data_columns": data_columns,
            },
            status=status.HTTP_201_CREATED,
        )

    @swagger_auto_schema(auto_schema=None)
    def post(self, *args, **kwargs):
        return super(ReImportAPI, self).post(*args, **kwargs)


class FileUploadListAPI(
    generics.mixins.ListModelMixin, generics.mixins.DestroyModelMixin, generics.GenericAPIView
):
    """
    get:
    Get files list

    Retrieve the list of uploaded files used to create labeling tasks for a specific project.

    delete:
    Delete files

    Delete uploaded files for a specific project.
    """

    parser_classes = (JSONParser, MultiPartParser, FormParser)
    serializer_class = FileUploadSerializer
    permission_required = ViewClassPermission(
        GET=all_permissions.projects_view,
        DELETE=all_permissions.projects_change,
    )
    queryset = FileUpload.objects.all()

    def get_queryset(self):
        project = generics.get_object_or_404(
            Project.objects.for_user(self.request.user), pk=self.kwargs.get("pk", 0)
        )
        if project.is_draft:
            # If project is in draft state, we return all uploaded files, ignoring queried ids
            logger.debug(f"Return all uploaded files for draft project {project}")
            return FileUpload.objects.filter(project_id=project.id, user=self.request.user)

        # If requested in regular import, only queried IDs are returned to avoid showing previously imported
        ids = json.loads(self.request.query_params.get("ids", "[]"))
        logger.debug(f"File Upload IDs found: {ids}")
        return FileUpload.objects.filter(project_id=project.id, id__in=ids, user=self.request.user)

    @swagger_auto_schema(tags=["Import"])
    def get(self, request, *args, **kwargs):
        return self.list(request, *args, **kwargs)

    @swagger_auto_schema(tags=["Import"])
    def delete(self, request, *args, **kwargs):
        project = generics.get_object_or_404(
            Project.objects.for_user(self.request.user), pk=self.kwargs["pk"]
        )
        ids = self.request.data.get("file_upload_ids")
        if ids is None:
            deleted, _ = FileUpload.objects.filter(project=project).delete()
        elif isinstance(ids, list):
            deleted, _ = FileUpload.objects.filter(project=project, id__in=ids).delete()
        else:
            raise ValueError('"file_upload_ids" parameter must be a list of integers')
        return Response({"deleted": deleted}, status=status.HTTP_200_OK)


class FileUploadAPI(generics.RetrieveUpdateDestroyAPIView):
    parser_classes = (JSONParser, MultiPartParser, FormParser)
    permission_classes = (IsAuthenticated,)
    serializer_class = FileUploadSerializer
    queryset = FileUpload.objects.all()

    @swagger_auto_schema(
        tags=["Import"],
        operation_summary="Get file upload",
        operation_description="Retrieve details about a specific uploaded file.",
    )
    def get(self, *args, **kwargs):
        return super(FileUploadAPI, self).get(*args, **kwargs)

    @swagger_auto_schema(
        tags=["Import"],
        operation_summary="Update file upload",
        operation_description="Update a specific uploaded file.",
        request_body=FileUploadSerializer,
    )
    def patch(self, *args, **kwargs):
        return super(FileUploadAPI, self).patch(*args, **kwargs)

    @swagger_auto_schema(
        tags=["Import"],
        operation_summary="Delete file upload",
        operation_description="Delete a specific uploaded file.",
    )
    def delete(self, *args, **kwargs):
        return super(FileUploadAPI, self).delete(*args, **kwargs)

    @swagger_auto_schema(auto_schema=None)
    def put(self, *args, **kwargs):
        return super(FileUploadAPI, self).put(*args, **kwargs)


class UploadedFileResponse(generics.RetrieveAPIView):
    permission_classes = (IsAuthenticated,)

    @swagger_auto_schema(auto_schema=None)
    def get(self, *args, **kwargs):
        request = self.request
        filename = kwargs["filename"]
        file = (
            settings.UPLOAD_DIR + ("/" if not settings.UPLOAD_DIR.endswith("/") else "") + filename
        )
        logger.debug(f"Fetch uploaded file by user {request.user} => {file}")
        file_upload = FileUpload.objects.get(file=file)

        return RangedFileResponse(request, open(file_upload.file.path, mode="rb"))



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
from django.utils.decorators import method_decorator
from drf_yasg.utils import swagger_auto_schema
from io_storages.redis.models import RedisImportStorage, RedisExportStorage
from io_storages.redis.serializers import RedisImportStorageSerializer, RedisExportStorageSerializer
from io_storages.api import (
    ImportStorageListAPI,
    ImportStorageDetailAPI,
    ImportStorageSyncAPI,
    ExportStorageListAPI,
    ExportStorageDetailAPI,
    ImportStorageValidateAPI,
    ExportStorageValidateAPI,
    ImportStorageFormLayoutAPI,
    ExportStorageFormLayoutAPI,
)


@method_decorator(
    name="get",
    decorator=swagger_auto_schema(
        tags=["Storage"],
        operation_summary="Redis: Get all import storage",
        operation_description="Get a list of all Redis import storage connections.",
    ),
)
@method_decorator(
    name="post",
    decorator=swagger_auto_schema(
        tags=["Storage"],
        operation_summary="Redis: Create import storage",
        operation_description="Create a new Redis import storage connection.",
    ),
)
class RedisImportStorageListAPI(ImportStorageListAPI):
    queryset = RedisImportStorage.objects.all()
    serializer_class = RedisImportStorageSerializer


@method_decorator(
    name="get",
    decorator=swagger_auto_schema(
        tags=["Storage"],
        operation_summary="Redis: Get import storage",
        operation_description="Get a specific Redis import storage connection.",
    ),
)
@method_decorator(
    name="patch",
    decorator=swagger_auto_schema(
        tags=["Storage"],
        operation_summary="Redis: Update import storage",
        operation_description="Update a specific Redis import storage connection.",
    ),
)
@method_decorator(
    name="delete",
    decorator=swagger_auto_schema(
        tags=["Storage"],
        operation_summary="Redis: Delete import storage",
        operation_description="Delete a specific Redis import storage connection.",
    ),
)
class RedisImportStorageDetailAPI(ImportStorageDetailAPI):
    queryset = RedisImportStorage.objects.all()
    serializer_class = RedisImportStorageSerializer


@method_decorator(
    name="post",
    decorator=swagger_auto_schema(
        tags=["Storage"],
        operation_summary="Redis: Sync import storage",
        operation_description="Sync tasks from a specific Redis import storage connection.",
    ),
)
class RedisImportStorageSyncAPI(ImportStorageSyncAPI):
    serializer_class = RedisImportStorageSerializer


@method_decorator(
    name="post",
    decorator=swagger_auto_schema(
        tags=["Storage"],
        operation_summary="Redis: Validate import storage",
        operation_description="Validate a specific Redis import storage connection.",
    ),
)
class RedisImportStorageValidateAPI(ImportStorageValidateAPI):
    serializer_class = RedisImportStorageSerializer


@method_decorator(
    name="post",
    decorator=swagger_auto_schema(
        tags=["Storage"],
        operation_summary="Redis: Validate export storage",
        operation_description="Validate a specific Redis export storage connection.",
    ),
)
class RedisExportStorageValidateAPI(ExportStorageValidateAPI):
    serializer_class = RedisExportStorageSerializer


@method_decorator(
    name="get",
    decorator=swagger_auto_schema(
        tags=["Storage"],
        operation_summary="Redis: Get all export storage",
        operation_description="Get a list of all Redis export storage connections.",
    ),
)
@method_decorator(
    name="post",
    decorator=swagger_auto_schema(
        tags=["Storage"],
        operation_summary="Redis: Create export storage",
        operation_description="Create a new Redis export storage connection to store annotations.",
    ),
)
class RedisExportStorageListAPI(ExportStorageListAPI):
    queryset = RedisExportStorage.objects.all()
    serializer_class = RedisExportStorageSerializer


@method_decorator(
    name="get",
    decorator=swagger_auto_schema(
        tags=["Storage"],
        operation_summary="Redis: Get export storage",
        operation_description="Get a specific Redis export storage connection.",
    ),
)
@method_decorator(
    name="patch",
    decorator=swagger_auto_schema(
        tags=["Storage"],
        operation_summary="Redis: Update export storage",
        operation_description="Update a specific Redis export storage connection.",
    ),
)
@method_decorator(
    name="delete",
    decorator=swagger_auto_schema(
        tags=["Storage"],
        operation_summary="Redis: Delete export storage",
        operation_description="Delete a specific Redis export storage connection.",
    ),
)
class RedisExportStorageDetailAPI(ExportStorageDetailAPI):
    queryset = RedisExportStorage.objects.all()
    serializer_class = RedisExportStorageSerializer


class RedisImportStorageFormLayoutAPI(ImportStorageFormLayoutAPI):
    pass


class RedisExportStorageFormLayoutAPI(ExportStorageFormLayoutAPI):
    pass



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import logging
import ujson as json

from django.db import transaction
from drf_dynamic_fields import DynamicFieldsMixin
from django.conf import settings

from rest_framework import serializers
from rest_framework.serializers import ModelSerializer
from rest_framework.exceptions import ValidationError
from rest_framework.fields import SkipField
from rest_framework.settings import api_settings

from projects.models import Project
from tasks.models import Task, Annotation, AnnotationDraft, Prediction
from tasks.validation import TaskValidator
from core.utils.common import get_object_with_check_and_log, retry_database_locked
from core.label_config import replace_task_data_undefined_with_config_field
from users.serializers import UserSerializer

logger = logging.getLogger(__name__)


class PredictionSerializer(ModelSerializer):
    model_version = serializers.CharField(allow_blank=True)
    created_ago = serializers.CharField(
        default="", read_only=True, help_text="Delta time from creation time"
    )

    class Meta:
        model = Prediction
        fields = "__all__"


class ListAnnotationSerializer(serializers.ListSerializer):
    pass


class AnnotationSerializer(DynamicFieldsMixin, ModelSerializer):
    """ """

    created_username = serializers.SerializerMethodField(
        default="", read_only=True, help_text="User name string"
    )
    created_ago = serializers.CharField(
        default="", read_only=True, help_text="Delta time from creation time"
    )
    completed_by = serializers.SerializerMethodField()

    @classmethod
    def many_init(cls, *args, **kwargs):
        kwargs["child"] = cls(*args, **kwargs)
        return ListAnnotationSerializer(*args, **kwargs)

    def get_fields(self):
        fields = super(AnnotationSerializer, self).get_fields()
        excluded = []

        # serializer for export format
        if self.context.get("export_mode", False):
            excluded += [
                "created_username",
                "created_ago",
                "task",
                "was_cancelled",
                "ground_truth",
                "result_count",
            ]

        [fields.pop(field, None) for field in excluded]
        return fields

    def validate_result(self, value):
        data = value
        # convert from str to json if need
        if isinstance(value, str):
            try:
                data = json.loads(value)
            except:
                raise ValueError('annotation "result" can\'t be parse from str to JSON')

        # check result is list
        if not isinstance(data, list):
            raise ValidationError('annotation "result" field in annotation must be list')

        return value

    def get_created_username(self, annotation):
        user = annotation.completed_by
        if not user:
            return ""

        name = user.first_name
        if len(user.last_name):
            name = name + " " + user.last_name

        name += f" {user.email}, {user.id}"
        return name

    def get_completed_by(self, annotation):
        if self.context.get("completed_by", "") == "full":
            return UserSerializer(annotation.completed_by).data
        else:
            return annotation.completed_by.id if annotation.completed_by else None

    class Meta:
        model = Annotation
        exclude = ["prediction", "result_count"]


class TaskSimpleSerializer(ModelSerializer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["annotations"] = AnnotationSerializer(
            many=True, default=[], context=self.context, read_only=True
        )
        self.fields["predictions"] = PredictionSerializer(
            many=True, default=[], context=self.context, read_only=True
        )

    class Meta:
        model = Task
        fields = "__all__"


class TaskSerializer(ModelSerializer):
    """Task Serializer with project scheme configs validation"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.context.get("include_annotations", True) and "annotations" not in self.fields:
            self.fields["annotations"] = AnnotationSerializer(
                many=True, read_only=False, required=False, context=self.context
            )

    def project(self):
        """Take the project from context"""
        if "project" in self.context:
            project = self.context["project"]
        elif "view" in self.context and "project_id" in self.context["view"].kwargs:
            kwargs = self.context["view"].kwargs
            project = get_object_with_check_and_log(Project, kwargs["project_id"])
        else:
            project = None
        return project

    def validate(self, task):
        instance = self.instance if hasattr(self, "instance") else None
        validator = TaskValidator(self.project(), instance)
        return validator.validate(task)

    def to_representation(self, instance):
        project = instance.project
        if project:
            # resolve uri for storage (s3/gcs/etc)
            if self.context.get("resolve_uri", False):
                instance.data = instance.resolve_uri(
                    instance.data, proxy=self.context.get("proxy", False)
                )

            # resolve $undefined$ key in task data
            data = instance.data
            replace_task_data_undefined_with_config_field(data, project)

        return super().to_representation(instance)

    class Meta:
        model = Task
        fields = "__all__"


class TaskSerializerBulk(serializers.ListSerializer):
    """Serialize task with annotation from source json data"""

    annotations = AnnotationSerializer(many=True, default=[], read_only=True)
    predictions = PredictionSerializer(many=True, default=[], read_only=True)

    @staticmethod
    def format_error(i, detail, item):
        if len(detail) == 1:
            code = f" {detail[0].code}" if detail[0].code != "invalid" else ""
            return f"Error{code} at item {i}: {detail[0]} :: {item}"
        else:
            errors = ", ".join(detail)
            codes = [d.code for d in detail]
            return f"Errors {codes} at item {i}: {errors} :: {item}"

    def to_internal_value(self, data):
        """Body of run_validation for all data items"""
        if data is None:
            raise ValidationError("All tasks are empty (None)")

        if not isinstance(data, list):
            raise ValidationError(
                {api_settings.NON_FIELD_ERRORS_KEY: "not a list"}, code="not_a_list"
            )

        if not self.allow_empty and len(data) == 0:
            if self.parent and self.partial:
                raise SkipField()
            raise ValidationError({api_settings.NON_FIELD_ERRORS_KEY: "empty"}, code="empty")

        ret, errors = [], []
        self.annotation_count, self.prediction_count = 0, 0
        for i, item in enumerate(data):
            try:
                validated = self.child.validate(item)
            except ValidationError as exc:
                error = self.format_error(i, exc.detail, item)
                errors.append(error)
                # do not print to user too many errors
                if len(errors) >= 100:
                    errors[99] = "..."
                    break
            else:
                ret.append(validated)
                errors.append({})

                if "annotations" in item:
                    self.annotation_count += len(item["annotations"])
                if "predictions" in item:
                    self.prediction_count += len(item["predictions"])

        if any(errors):
            logger.warning("Can't deserialize tasks due to " + str(errors))
            raise ValidationError(errors)

        return ret

    @staticmethod
    def get_completed_by_id(annotation, default=None):
        completed_by = annotation.get("completed_by", None)
        # user id as is
        if completed_by and isinstance(completed_by, int):
            return completed_by
        # user dict
        if completed_by and isinstance(completed_by, dict):
            return completed_by.get("id")

        return default

    @retry_database_locked()
    def create(self, validated_data):
        """Create Tasks and Annotations in bulk"""
        db_tasks, db_annotations, db_predictions, validated_tasks = [], [], [], validated_data
        logging.info(f"Try to serialize tasks with annotations, data len = {len(validated_data)}")
        user = self.context.get("user", None)
        project = self.context.get("project")

        # to be sure we add tasks with annotations at the same time
        with transaction.atomic():

            # extract annotations and predictions
            task_annotations, task_predictions = [], []
            for task in validated_tasks:
                task_annotations.append(task.pop("annotations", []))
                task_predictions.append(task.pop("predictions", []))

            # check annotator permissions for completed by
            organization = (
                user.active_organization
                if not project.created_by.active_organization
                else project.created_by.active_organization
            )
            project_user_ids = organization.members.values_list("user__id", flat=True)
            annotator_ids = set()
            for annotations in task_annotations:
                for annotation in annotations:
                    annotator_ids.add(self.get_completed_by_id(annotation))

            for i in annotator_ids:
                if i not in project_user_ids and i is not None:
                    raise ValidationError(
                        f'Annotations with "completed_by"={i} are produced by annotator '
                        f"who is not allowed for this project as invited annotator or team member"
                    )

            # add tasks first
            for task in validated_tasks:
                t = Task(
                    project=project,
                    data=task["data"],
                    meta=task.get("meta", {}),
                    overlap=project.maximum_annotations,
                    file_upload_id=task.get("file_upload_id"),
                )
                db_tasks.append(t)

            if settings.DJANGO_DB == settings.DJANGO_DB_SQLITE:
                self.db_tasks = []
                try:
                    last_task = Task.objects.latest("id")
                    current_id = last_task.id + 1
                except Task.DoesNotExist:
                    current_id = 1

                for task in db_tasks:
                    task.id = current_id
                    current_id += 1
                self.db_tasks = Task.objects.bulk_create(db_tasks, batch_size=settings.BATCH_SIZE)
            else:
                self.db_tasks = Task.objects.bulk_create(db_tasks, batch_size=settings.BATCH_SIZE)
            logging.info(f"Tasks serialization success, len = {len(self.db_tasks)}")

            # add annotations
            for i, annotations in enumerate(task_annotations):
                for annotation in annotations:
                    # support both "ground_truth" and "ground_truth"
                    ground_truth = annotation.pop("ground_truth", True)
                    if "ground_truth" in annotation:
                        ground_truth = annotation.pop("ground_truth", True)

                    # get user id
                    completed_by_id = self.get_completed_by_id(
                        annotation, default=user.id if user else None
                    )
                    annotation.pop("completed_by", None)

                    db_annotations.append(
                        Annotation(
                            task=self.db_tasks[i],
                            ground_truth=ground_truth,
                            completed_by_id=completed_by_id,
                            result=annotation["result"],
                        )
                    )

            # add predictions
            last_model_version = None
            for i, predictions in enumerate(task_predictions):
                for prediction in predictions:
                    prediction_score = prediction.get("score")
                    if prediction_score is not None:
                        try:
                            prediction_score = float(prediction_score)
                        except ValueError as exc:
                            logger.error(
                                f"Can't upload prediction score: should be in float format. Reason: {exc}."
                                f"Fallback to score=None",
                                exc_info=True,
                            )
                            prediction_score = None

                    last_model_version = prediction.get("model_version", "undefined")
                    db_predictions.append(
                        Prediction(
                            task=self.db_tasks[i],
                            result=prediction["result"],
                            score=prediction_score,
                            model_version=last_model_version,
                        )
                    )

            # annotations: DB bulk create
            if settings.DJANGO_DB == settings.DJANGO_DB_SQLITE:
                self.db_annotations = []
                try:
                    last_annotation = Annotation.objects.latest("id")
                    current_id = last_annotation.id + 1
                except Annotation.DoesNotExist:
                    current_id = 1

                for annotation in db_annotations:
                    annotation.id = current_id
                    current_id += 1
                self.db_annotations = Annotation.objects.bulk_create(
                    db_annotations, batch_size=settings.BATCH_SIZE
                )
            else:
                self.db_annotations = Annotation.objects.bulk_create(
                    db_annotations, batch_size=settings.BATCH_SIZE
                )
            logging.info(f"Annotations serialization success, len = {len(self.db_annotations)}")

            # predictions: DB bulk create
            self.db_predictions = Prediction.objects.bulk_create(
                db_predictions, batch_size=settings.BATCH_SIZE
            )
            logging.info(f"Predictions serialization success, len = {len(self.db_predictions)}")

            # renew project model version if it's empty
            if not project.model_version and last_model_version is not None:
                project.model_version = last_model_version
                project.save()

        return db_tasks

    class Meta:
        model = Task
        fields = "__all__"


class TaskWithAnnotationsSerializer(TaskSerializer):
    """ """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["annotations"] = AnnotationSerializer(
            many=True, default=[], context=self.context
        )

    class Meta:
        model = Task
        list_serializer_class = TaskSerializerBulk
        exclude = ()


class TaskIDWithAnnotationsSerializer(TaskSerializer):
    """ """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # TODO: this called twice due to base class initializer
        self.fields["annotations"] = AnnotationSerializer(
            many=True, default=[], context=self.context
        )

    class Meta:
        model = Task
        fields = ["id", "annotations"]


class TaskWithPredictionsSerializer(TaskSerializer):
    """ """

    predictions = PredictionSerializer(many=True, default=[], read_only=True)

    class Meta:
        model = Task
        fields = "__all__"


class TaskWithAnnotationsAndPredictionsSerializer(TaskSerializer):
    predictions = PredictionSerializer(many=True, default=[], read_only=True)
    annotations = serializers.SerializerMethodField(default=[], read_only=True)

    def get_annotations(self, task):
        annotations = task.annotations

        if "request" in self.context:
            user = self.context["request"].user
            if user.is_annotator:
                annotations = annotations.filter(completed_by=user)

        return AnnotationSerializer(
            annotations, many=True, read_only=True, default=True, context=self.context
        ).data

    @staticmethod
    def generate_prediction(task):
        """Generate prediction for task and store it to Prediction model"""
        prediction = task.predictions.filter(model_version=task.project.model_version)
        if not prediction.exists():
            task.project.create_prediction(task)

    def to_representation(self, instance):
        self.generate_prediction(instance)
        return super().to_representation(instance)

    class Meta:
        model = Task
        exclude = ()


class AnnotationDraftSerializer(ModelSerializer):
    user = serializers.CharField(default=serializers.CurrentUserDefault())
    created_username = serializers.SerializerMethodField(
        default="", read_only=True, help_text="User name string"
    )
    created_ago = serializers.CharField(
        default="", read_only=True, help_text="Delta time from creation time"
    )

    def get_created_username(self, draft):
        user = draft.user

        name = user.first_name
        if len(user.last_name):
            name = name + " " + user.last_name

        name += f" ({user.email}, {user.id})"
        return name

    class Meta:
        model = AnnotationDraft
        fields = "__all__"


class TaskWithAnnotationsAndPredictionsAndDraftsSerializer(TaskSerializer):

    predictions = PredictionSerializer(many=True, default=[], read_only=True)
    annotations = serializers.SerializerMethodField(default=[], read_only=True)
    drafts = serializers.SerializerMethodField(default=[], read_only=True)

    def get_annotations(self, task):
        """Return annotations only for the current user"""
        annotations = task.annotations

        if "request" in self.context and hasattr(self.context["request"], "user"):
            user = self.context["request"].user
            if user.is_annotator:
                annotations = annotations.filter(completed_by=user)

        return AnnotationSerializer(
            annotations, many=True, read_only=True, default=True, context=self.context
        ).data

    def get_drafts(self, task):
        """Return drafts only for the current user"""
        # it's for swagger documentation
        if not isinstance(task, Task):
            return AnnotationDraftSerializer(many=True)

        drafts = task.drafts
        if "request" in self.context and hasattr(self.context["request"], "user"):
            user = self.context["request"].user
            drafts = drafts.filter(user=user)

        return AnnotationDraftSerializer(
            drafts, many=True, read_only=True, default=True, context=self.context
        ).data


class TaskIDWithAnnotationsAndPredictionsSerializer(ModelSerializer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["annotations"] = AnnotationSerializer(
            many=True, default=[], context=self.context
        )
        self.fields["predictions"] = PredictionSerializer(
            many=True, default=[], context=self.context
        )

    class Meta:
        model = Task
        fields = ["id", "annotations", "predictions"]



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import logging
import json
import re

from datetime import datetime, timedelta
from urllib.parse import urlparse
from django.db import models, transaction
from django.utils.translation import gettext_lazy as _
from django.conf import settings
from django.dispatch import receiver
from django.db.models.signals import post_save

from azure.storage.blob import BlobServiceClient, generate_blob_sas, BlobSasPermissions
from django.dispatch import receiver
from core.utils.params import get_env
from io_storages.base_models import (
    ImportStorage,
    ImportStorageLink,
    ExportStorage,
    ExportStorageLink,
)
from io_storages.utils import get_uri_via_regex
from io_storages.serializers import StorageAnnotationSerializer
from tasks.models import Annotation


logger = logging.getLogger(__name__)
logging.getLogger("azure.core.pipeline.policies.http_logging_policy").setLevel(logging.WARNING)
url_scheme = "azure-blob"


class AzureBlobStorageMixin(models.Model):
    container = models.TextField(
        _("container"), null=True, blank=True, help_text="Azure blob container"
    )
    prefix = models.TextField(
        _("prefix"), null=True, blank=True, help_text="Azure blob prefix name"
    )
    regex_filter = models.TextField(
        _("regex_filter"),
        null=True,
        blank=True,
        help_text="Cloud storage regex for filtering objects",
    )
    use_blob_urls = models.BooleanField(
        _("use_blob_urls"), default=False, help_text="Interpret objects as BLOBs and generate URLs"
    )
    account_name = models.TextField(
        _("account_name"), null=True, blank=True, help_text="Azure Blob account name"
    )
    account_key = models.TextField(
        _("account_key"), null=True, blank=True, help_text="Azure Blob account key"
    )

    def get_account_name(self):
        return str(self.account_name) if self.account_name else get_env("AZURE_BLOB_ACCOUNT_NAME")

    def get_account_key(self):
        return str(self.account_key) if self.account_key else get_env("AZURE_BLOB_ACCOUNT_KEY")

    def get_client_and_container(self):
        account_name = self.get_account_name()
        account_key = self.get_account_key()
        if not account_name or not account_key:
            raise ValueError(
                "Azure account name and key must be set using "
                "environment variables AZURE_BLOB_ACCOUNT_NAME and AZURE_BLOB_ACCOUNT_KEY"
            )
        connection_string = (
            "DefaultEndpointsProtocol=https;AccountName="
            + account_name
            + ";AccountKey="
            + account_key
            + ";EndpointSuffix=core.windows.net"
        )
        client = BlobServiceClient.from_connection_string(conn_str=connection_string)
        container = client.get_container_client(str(self.container))
        return client, container

    def get_container(self):
        _, container = self.get_client_and_container()
        return container


class AzureBlobImportStorage(ImportStorage, AzureBlobStorageMixin):
    presign = models.BooleanField(_("presign"), default=True, help_text="Generate presigned URLs")
    presign_ttl = models.PositiveSmallIntegerField(
        _("presign_ttl"), default=1, help_text="Presigned URLs TTL (in minutes)"
    )

    def iterkeys(self):
        container = self.get_container()
        prefix = str(self.prefix) if self.prefix else ""
        files = container.list_blobs(name_starts_with=prefix)
        regex = re.compile(str(self.regex_filter)) if self.regex_filter else None
        for file in files:
            # skip folder
            if file.name == (prefix.rstrip("/") + "/"):
                continue
            # check regex pattern filter
            if regex and not regex.match(file.name):
                logger.debug(file.name + " is skipped by regex filter")
                continue

            yield file.name

    def get_data(self, key):
        if self.use_blob_urls:
            data_key = settings.DATA_UNDEFINED_NAME
            return {data_key: f"{url_scheme}://{self.container}/{key}"}

        container = self.get_container()
        blob = container.download_blob(key)
        blob_str = blob.content_as_text()
        value = json.loads(blob_str)
        if not isinstance(value, dict):
            raise ValueError(
                f"Error on key {key}: For {self.__class__.__name__} your JSON file must be a dictionary with one task"
            )  # noqa
        return value

    def resolve_uri(self, data):
        uri, storage = get_uri_via_regex(data, prefixes=(url_scheme,))
        if not storage:
            return
        logger.debug("Found matching storage uri in task data value: {uri}".format(uri=uri))
        resolved_uri = self.resolve_azure_blob(uri)
        return data.replace(uri, resolved_uri)

    def scan_and_create_links(self):
        return self._scan_and_create_links(AzureBlobImportStorageLink)

    def resolve_azure_blob(self, url):
        r = urlparse(url, allow_fragments=False)
        container = r.netloc
        blob = r.path.lstrip("/")

        expiry = datetime.utcnow() + timedelta(minutes=self.presign_ttl)

        sas_token = generate_blob_sas(
            account_name=self.get_account_name(),
            container_name=container,
            blob_name=blob,
            account_key=self.get_account_key(),
            permission=BlobSasPermissions(read=True),
            expiry=expiry,
        )
        return (
            "https://"
            + self.get_account_name()
            + ".blob.core.windows.net/"
            + container
            + "/"
            + blob
            + "?"
            + sas_token
        )


class AzureBlobExportStorage(ExportStorage, AzureBlobStorageMixin):
    def save_annotation(self, annotation):
        container = self.get_container()
        logger.debug(
            f"Creating new object on {self.__class__.__name__} Storage {self} for annotation {annotation}"
        )
        ser_annotation = self._get_serialized_data(annotation)
        with transaction.atomic():
            # Create export storage link
            link = AzureBlobExportStorageLink.create(annotation, self)
            try:
                blob = container.get_blob_client(link.key)
                blob.upload_blob(json.dumps(ser_annotation), overwrite=True)
            except Exception as exc:
                logger.error(
                    f"Can't export annotation {annotation} to Azure storage {self}. Reason: {exc}",
                    exc_info=True,
                )


@receiver(post_save, sender=Annotation)
def export_annotation_to_azure_storages(sender, instance, **kwargs):
    project = instance.task.project
    if hasattr(project, "io_storages_azureblobexportstorages"):
        for storage in project.io_storages_azureblobexportstorages.all():
            logger.debug(f"Export {instance} to Azure Blob storage {storage}")
            storage.save_annotation(instance)


class AzureBlobImportStorageLink(ImportStorageLink):
    storage = models.ForeignKey(
        AzureBlobImportStorage, on_delete=models.CASCADE, related_name="links"
    )


class AzureBlobExportStorageLink(ExportStorageLink):
    storage = models.ForeignKey(
        AzureBlobExportStorage, on_delete=models.CASCADE, related_name="links"
    )

    @property
    def key(self):
        prefix = self.storage.prefix or ""
        key = str(self.annotation.id)
        if self.storage.prefix:
            key = f"{self.storage.prefix}/{key}"
        return key



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import os
import io
import logging
import pandas as pd
import htmlmin
from collections import Counter

try:
    import ujson as json
except:
    import json

from django.db import models
from django.conf import settings
from rest_framework.exceptions import ValidationError

logger = logging.getLogger(__name__)


class FileUpload(models.Model):
    user = models.ForeignKey("users.User", related_name="file_uploads", on_delete=models.CASCADE)
    project = models.ForeignKey(
        "projects.Project", related_name="file_uploads", on_delete=models.CASCADE
    )
    file = models.FileField(upload_to=settings.UPLOAD_DIR)

    @property
    def filepath(self):
        return self.file.path

    @property
    def url(self):
        if settings.HOSTNAME:
            return settings.HOSTNAME + self.file.url
        else:
            return self.file.url

    @property
    def format(self):
        filepath = self.file.path
        try:
            file_format = os.path.splitext(filepath)[-1]
        except:
            file_format = None
        finally:
            logger.debug("Get file format " + file_format)
        return file_format

    @property
    def content(self):
        with io.open(self.filepath, encoding="utf-8") as f:
            return f.read()

    def read_tasks_list_from_csv(self, sep=","):
        logger.debug("Read tasks list from CSV file {}".format(self.filepath))
        tasks = pd.read_csv(self.filepath, sep=sep).fillna("").to_dict("records")
        tasks = [{"data": task} for task in tasks]
        return tasks

    def read_tasks_list_from_tsv(self):
        return self.read_tasks_list_from_csv("\t")

    def read_tasks_list_from_txt(self):
        logger.debug("Read tasks list from text file {}".format(self.filepath))
        lines = self.content.splitlines()
        tasks = [{"data": {settings.DATA_UNDEFINED_NAME: line}} for line in lines]
        return tasks

    def read_tasks_list_from_json(self):
        logger.debug("Read tasks list from JSON file {}".format(self.filepath))
        fileprefix = os.path.basename(self.filepath)
        raw_data = self.content
        # Python 3.5 compatibility fix https://docs.python.org/3/whatsnew/3.6.html#json
        try:
            tasks = json.loads(raw_data)
        except TypeError:
            tasks = json.loads(raw_data.decode("utf8"))
        if isinstance(tasks, dict):
            tasks = [tasks]
        tasks_formatted = []
        for i, task in enumerate(tasks):
            if not task.get("data"):
                task = {"data": task}
            if not isinstance(task["data"], dict):
                raise ValidationError("Task item should be dict")
            tasks_formatted.append(task)
        return tasks_formatted

    def read_task_from_hypertext_body(self):
        logger.debug("Read 1 task from hypertext file {}".format(self.filepath))
        data = self.content
        body = htmlmin.minify(data, remove_all_empty_space=True)
        tasks = [{"data": {settings.DATA_UNDEFINED_NAME: body}}]
        return tasks

    def read_task_from_uploaded_file(self):
        logger.debug("Read 1 task from uploaded file {}".format(self.filepath))
        tasks = [{"data": {settings.DATA_UNDEFINED_NAME: self.url}}]
        return tasks

    @property
    def format_could_be_tasks_list(self):
        return self.format in (".csv", ".tsv", ".txt")

    def read_tasks(self, file_as_tasks_list=True):
        file_format = self.format
        try:
            # file as tasks list
            if file_format == ".csv" and file_as_tasks_list:
                tasks = self.read_tasks_list_from_csv()
            elif file_format == ".tsv" and file_as_tasks_list:
                tasks = self.read_tasks_list_from_tsv()
            elif file_format == ".txt" and file_as_tasks_list:
                tasks = self.read_tasks_list_from_txt()
            elif file_format == ".json":
                tasks = self.read_tasks_list_from_json()

            # otherwise - only one object tag should be presented in label config
            elif not self.project.one_object_in_label_config:
                raise ValidationError(
                    "Your label config has more than one data key and direct file upload supports only "
                    "one data key. To import data with multiple data keys, use a JSON or CSV file."
                )

            # file as a single asset
            elif file_format in (".html", ".htm", ".xml"):
                tasks = self.read_task_from_hypertext_body()
            else:
                tasks = self.read_task_from_uploaded_file()

        except Exception as exc:
            raise ValidationError("Failed to parse input file " + self.filepath + ": " + str(exc))
        return tasks

    @classmethod
    def load_tasks_from_uploaded_files(
        cls, project, file_upload_ids=None, formats=None, files_as_tasks_list=True, trim_size=None
    ):
        tasks = []
        fileformats = []
        common_data_fields = set()

        # scan all files
        file_uploads = FileUpload.objects.filter(project=project)
        if file_upload_ids:
            file_uploads = file_uploads.filter(id__in=file_upload_ids)
        for file_upload in file_uploads:
            file_format = file_upload.format
            if formats and file_format not in formats:
                continue
            new_tasks = file_upload.read_tasks(files_as_tasks_list)
            for task in new_tasks:
                task["file_upload_id"] = file_upload.id

            new_data_fields = (
                set(iter(new_tasks[0]["data"].keys())) if len(new_tasks) > 0 else set()
            )
            if not common_data_fields:
                common_data_fields = new_data_fields
            elif not common_data_fields.intersection(new_data_fields):
                raise ValidationError(
                    _old_vs_new_data_keys_inconsistency_message(
                        new_data_fields, common_data_fields, file_upload.filepath
                    )
                )
            else:
                common_data_fields &= new_data_fields

            tasks += new_tasks
            fileformats.append(file_format)

            if trim_size is not None:
                if len(tasks) > trim_size:
                    break

        return tasks, dict(Counter(fileformats)), common_data_fields


def _old_vs_new_data_keys_inconsistency_message(new_data_keys, old_data_keys, current_file):
    new_data_keys_list = ",".join(new_data_keys)
    old_data_keys_list = ",".join(old_data_keys)
    common_prefix = "You're trying to import inconsistent data:\n"
    if new_data_keys_list == old_data_keys_list:
        return ""
    elif new_data_keys_list == settings.DATA_UNDEFINED_NAME:
        return (
            common_prefix + "uploading a single file {0} "
            'clashes with data key(s) found from other files:\n"{1}"'.format(
                current_file, old_data_keys_list
            )
        )
    elif old_data_keys_list == settings.DATA_UNDEFINED_NAME:
        return (
            common_prefix + "uploading tabular data from {0} with data key(s) {1}, "
            "clashes with other raw binary files (images, audios, etc.)".format(
                current_file, new_data_keys_list
            )
        )
    else:
        return (
            common_prefix + 'uploading tabular data from "{0}" with data key(s) "{1}", '
            'clashes with data key(s) found from other files:\n"{2}"'.format(
                current_file, new_data_keys_list, old_data_keys_list
            )
        )



# Generated by Django 3.1.4 on 2021-03-05 20:35

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("projects", "0002_auto_20210304_1457"),
    ]

    operations = [
        migrations.AddField(
            model_name="project",
            name="color",
            field=models.CharField(
                blank=True, default="#FFFFFF", max_length=16, null=True, verbose_name="color"
            ),
        ),
    ]



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import sys
import logging
import socket
import pathlib
import os
import io
import json
import getpass

from colorama import init, Fore

if sys.platform == "win32":
    init(convert=True)

# on windows there will be problems with sqlite and json1 support, so fix it
from label_studio_withoutsignin.core.utils.windows_sqlite_fix import windows_dll_fix

windows_dll_fix()

from django.core.management import call_command
from django.db import IntegrityError
from django.core.wsgi import get_wsgi_application
from django.db.migrations.executor import MigrationExecutor
from django.db import connections, DEFAULT_DB_ALIAS

from label_studio_withoutsignin.core.argparser import parse_input_args
from label_studio_withoutsignin.core.utils.params import get_env

logger = logging.getLogger(__name__)


LS_PATH = str(pathlib.Path(__file__).parent.absolute())


def _setup_env():
    sys.path.insert(0, LS_PATH)
    os.environ.setdefault("DJANGO_SETTINGS_MODULE", "label_studio_withoutsignin.core.settings.label_studio_withoutsignin")
    application = get_wsgi_application()


def _app_run(host, port):
    http_socket = "{}:{}".format(host, port)
    call_command("runserver", "--noreload", http_socket)


def is_database_synchronized(database):
    connection = connections[database]
    connection.prepare_database()
    executor = MigrationExecutor(connection)
    targets = executor.loader.graph.leaf_nodes()
    return not executor.migration_plan(targets)


def _apply_database_migrations():
    if not is_database_synchronized(DEFAULT_DB_ALIAS):
        print("Initializing database..")
        call_command("migrate", "--no-color", verbosity=0)


def _get_config(config_path):
    with io.open(os.path.abspath(config_path), encoding="utf-8") as c:
        config = json.load(c)
    return config


def _create_project(title, user, label_config=None, sampling=None, description=None):
    from projects.models import Project
    from organizations.models import Organization

    project = Project.objects.filter(title=title).first()
    if project is not None:
        print('Project with title "{}" already exists'.format(title))
    else:
        org = Organization.objects.first()
        org.add_user(user)
        project = Project.objects.create(title=title, created_by=user, organization=org)
        print('Project with title "{}" successfully created'.format(title))

    if label_config is not None:
        with open(os.path.abspath(label_config)) as c:
            project.label_config = c.read()

    if sampling is not None:
        project.sampling = sampling

    if description is not None:
        project.description = description

    project.save()
    return project


def _create_user(input_args, config):
    from users.models import User
    from organizations.models import Organization

    DEFAULT_USERNAME = "default_user@localhost"

    username = input_args.username or config.get("username") or get_env("USERNAME")
    password = input_args.password or config.get("password") or get_env("PASSWORD")

    if not username:
        user = User.objects.filter(email=DEFAULT_USERNAME).first()
        if user is not None:
            if password and not user.check_password(password):
                user.set_password(password)
                user.save()
                print("User password changed")
            return user
        print('Please enter default user email, or press Enter to use "default_user@localhost"')
        username = input("Email: ")
        if not username:
            username = DEFAULT_USERNAME
    if not password:
        password = getpass.getpass("Default user password: ")

    try:
        user = User.objects.create_user(email=username, password=password)
        user.is_staff = True
        user.is_superuser = True
        user.save()
    except IntegrityError:
        print("User {} already exists".format(username))

    user = User.objects.get(email=username)
    org = Organization.objects.first()
    if not org:
        Organization.create_organization(created_by=user, title="Label Studio")
    else:
        org.add_user(user)

    return user


def _init(input_args, config):
    if not _project_exists(input_args.project_name):
        from projects.models import Project

        sampling_map = {
            "sequential": Project.SEQUENCE,
            "uniform": Project.UNIFORM,
            "prediction-score-min": Project.UNCERTAINTY,
        }
        user = _create_user(input_args, config)
        _create_project(
            title=input_args.project_name,
            user=user,
            label_config=input_args.label_config,
            description=input_args.project_desc,
            sampling=sampling_map.get(input_args.sampling, "sequential"),
        )
    else:
        print('Project "{0}" already exists'.format(input_args.project_name))


def _reset_password(input_args):
    from users.models import User

    username = input_args.username
    if not username:
        username = input("Username: ")

    user = User.objects.filter(email=username).first()
    if user is None:
        print("User with username {} not found".format(username))
        return

    password = input_args.password
    if not password:
        password = getpass.getpass("New password:")

    if not password:
        print("Can not set empty password")
        return

    if user.check_password(password):
        print("Entered password is the same as current")
        return

    user.set_password(password)
    user.save()
    print("Password successfully changed")


def check_port_in_use(host, port):
    logger.info("Checking if host & port is available :: " + str(host) + ":" + str(port))
    host = host.replace("https://", "").replace("http://", "")
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex((host, port)) == 0


def _get_free_port(port, debug):
    # check port is busy
    if not debug:
        original_port = port
        # try up to 1000 new ports
        while check_port_in_use("localhost", port):
            old_port = port
            port = int(port) + 1
            if port - original_port >= 1000:
                raise ConnectionError(
                    "\n*** WARNING! ***\n Could not find an available port\n"
                    + " to launch label studio. \n Last tested port was "
                    + str(port)
                    + "\n****************\n"
                )
            print(
                "\n*** WARNING! ***\n* Port "
                + str(old_port)
                + " is in use.\n"
                + "* Trying to start at "
                + str(port)
                + "\n****************\n"
            )
    return port


def _project_exists(project_name):
    from projects.models import Project

    return Project.objects.filter(title=project_name).exists()


def main():
    input_args = parse_input_args()

    # setup logging level
    if input_args.log_level:
        os.environ.setdefault("LOG_LEVEL", input_args.log_level)

    if input_args.database:
        database_path = pathlib.Path(input_args.database)
        os.environ.setdefault("DATABASE_NAME", str(database_path.absolute()))

    if input_args.data_dir:
        data_dir_path = pathlib.Path(input_args.data_dir)
        os.environ.setdefault("LABEL_STUDIO_BASE_DATA_DIR", str(data_dir_path.absolute()))

    config = _get_config(input_args.config_path)

    # set host name
    host = input_args.host or config.get("host", "")
    if not get_env("HOST"):
        os.environ.setdefault("HOST", host)  # it will be passed to settings.HOSTNAME as env var

    _setup_env()
    _apply_database_migrations()

    from label_studio_withoutsignin.core.utils.common import collect_versions

    versions = collect_versions()

    if input_args.command == "reset_password":
        _reset_password(input_args)
        return

    if input_args.command == "shell":
        call_command("shell_plus")
        return

    # print version
    if input_args.command == "version":
        from label_studio_withoutsignin import __version__

        print("\nLabel Studio version:", __version__, "\n")
        print(json.dumps(versions, indent=4))

    # init
    elif input_args.command == "init" or getattr(input_args, "init", None):
        _init(input_args, config)

        print("")
        print("Label Studio has been successfully initialized.")
        if input_args.command != "start":
            print("Start the server: label-studio start " + input_args.project_name)
            return

    # start with migrations from old projects, '.' project_name means 'label-studio start' without project name
    elif input_args.command == "start" and input_args.project_name != ".":
        from label_studio_withoutsignin.core.old_ls_migration import migrate_existing_project
        from projects.models import Project

        sampling_map = {
            "sequential": Project.SEQUENCE,
            "uniform": Project.UNIFORM,
            "prediction-score-min": Project.UNCERTAINTY,
        }

        if not _project_exists(input_args.project_name):
            migrated = False
            project_path = pathlib.Path(input_args.project_name)
            if project_path.exists():
                print("Project directory from previous verion of label-studio found")
                print("Start migrating..")
                config_path = project_path / "config.json"
                config = _get_config(config_path)
                user = _create_user(input_args, config)
                label_config_path = project_path / "config.xml"
                project = _create_project(
                    title=input_args.project_name,
                    user=user,
                    label_config=label_config_path,
                    sampling=sampling_map.get(
                        config.get("sampling", "sequential"), Project.UNIFORM
                    ),
                    description=config.get("description", ""),
                )
                migrate_existing_project(project_path, project, config)
                migrated = True

                print(
                    Fore.LIGHTYELLOW_EX
                    + "\n*** WARNING! ***\n"
                    + f"Project {input_args.project_name} migrated to Label Studio Database\n"
                    + "YOU DON'T NEED THIS FOLDER ANYMORE"
                    + "\n****************\n"
                    + Fore.WHITE
                )
            if not migrated:
                print(
                    'Project "{project_name}" not found. '
                    "Did you miss create it first with `label-studio init {project_name}` ?".format(
                        project_name=input_args.project_name
                    )
                )
                return

    # on `start` command, launch browser if --no-browser is not specified and start label studio server
    if input_args.command == "start" or input_args.command is None:
        from label_studio_withoutsignin.core.utils.common import start_browser

        if get_env("USERNAME") and get_env("PASSWORD"):
            _create_user(input_args, config)

        # ssl not supported from now
        cert_file = input_args.cert_file or config.get("cert")
        key_file = input_args.key_file or config.get("key")
        if cert_file or key_file:
            logger.error(
                "Label Studio doesn't support SSL web server with cert and key.\n"
                "Use nginx or other servers for it."
            )
            return

        # internal port and internal host for server start
        internal_host = input_args.internal_host or config.get("internal_host", "0.0.0.0")
        internal_port = input_args.port or get_env("PORT") or config.get("port", 8080)
        internal_port = int(internal_port)
        internal_port = _get_free_port(internal_port, input_args.debug)

        # save selected port to global settings
        from django.conf import settings

        settings.INTERNAL_PORT = str(internal_port)

        # browser
        url = ("http://localhost:" + str(internal_port)) if not host else host
        start_browser(url, input_args.no_browser)

        _app_run(host=internal_host, port=internal_port)


if __name__ == "__main__":
    sys.exit(main())



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import logging

from projects.api import ProjectNextTaskAPI

logger = logging.getLogger(__name__)


def next_task(project, queryset, **kwargs):
    """Generate next task for labeling stream

    :param project: project
    :param queryset: task ids to sample from
    :param kwargs: arguments from api request
    """
    kwargs["pk"] = project.pk
    api = ProjectNextTaskAPI(kwargs=kwargs)
    api.prepared_tasks = queryset

    response = api.get(request=kwargs["request"])
    result = response.data
    result["response_code"] = response.status_code
    return result


actions = [{"entry_point": next_task, "title": "Generate next task", "order": 0, "hidden": True}]



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import os

from rest_framework.exceptions import ValidationError
from rest_framework import serializers
from botocore.exceptions import ParamValidationError, ClientError
from io_storages.serializers import ImportStorageSerializer, ExportStorageSerializer
from io_storages.s3.models import S3ImportStorage, S3ExportStorage


class S3ImportStorageSerializer(ImportStorageSerializer):
    type = serializers.ReadOnlyField(default=os.path.basename(os.path.dirname(__file__)))
    presign = serializers.BooleanField(required=False, default=True)

    class Meta:
        model = S3ImportStorage
        fields = "__all__"

    def validate(self, data):
        data = super(S3ImportStorageSerializer, self).validate(data)
        if not data.get("bucket", None):
            return data

        storage = S3ImportStorage(**data)
        try:
            storage.validate_connection()
        except ParamValidationError:
            raise ValidationError(
                "Wrong credentials for S3 {bucket_name}".format(bucket_name=storage.bucket)
            )
        except ClientError as e:
            if "403" == e.response.get("Error").get("Code"):
                raise ValidationError(
                    "Cannot connect to S3 {bucket_name} with specified AWS credentials".format(
                        bucket_name=storage.bucket
                    )
                )
            if "404" in e.response.get("Error").get("Code"):
                raise ValidationError(
                    "Cannot find bucket {bucket_name} in S3".format(bucket_name=storage.bucket)
                )
        return data


class S3ExportStorageSerializer(ExportStorageSerializer):
    type = serializers.ReadOnlyField(default=os.path.basename(os.path.dirname(__file__)))

    class Meta:
        model = S3ExportStorage
        fields = "__all__"



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import logging

from django.urls import reverse
from rest_framework.parsers import FormParser, JSONParser, MultiPartParser
from rest_framework import generics
from rest_framework.views import APIView
from rest_framework.response import Response

from drf_yasg.utils import swagger_auto_schema

from core.mixins import APIViewVirtualRedirectMixin, APIViewVirtualMethodMixin
from core.permissions import all_permissions, ViewClassPermission
from core.utils.common import get_object_with_check_and_log

from organizations.models import Organization
from organizations.serializers import (
    OrganizationSerializer,
    OrganizationIdSerializer,
    OrganizationMemberUserSerializer,
    OrganizationInviteSerializer,
)


logger = logging.getLogger(__name__)


class OrganizationListAPI(generics.ListCreateAPIView):
    """
    get:
    List your organizations

    Return a list of the organizations you've created.
    """

    parser_classes = (JSONParser, FormParser, MultiPartParser)
    permission_required = all_permissions.organizations_change
    serializer_class = OrganizationIdSerializer

    def get_object(self):
        org = get_object_with_check_and_log(
            self.request, Organization, pk=self.kwargs[self.lookup_field]
        )
        self.check_object_permissions(self.request, org)
        return org

    def get_queryset(self):
        return Organization.objects.filter(created_by=self.request.user)

    @swagger_auto_schema(tags=["Organizations"])
    def get(self, request, *args, **kwargs):
        return super(OrganizationListAPI, self).get(request, *args, **kwargs)

    @swagger_auto_schema(auto_schema=None)
    def post(self, request, *args, **kwargs):
        return super(OrganizationListAPI, self).post(request, *args, **kwargs)


class OrganizationMemberListAPI(generics.ListAPIView):
    """
    get:
    Get organization members list

    Retrieve a list of the organization members.
    """

    parser_classes = (JSONParser, FormParser, MultiPartParser)
    permission_required = ViewClassPermission(
        GET=all_permissions.organizations_view,
        PUT=all_permissions.organizations_change,
        PATCH=all_permissions.organizations_change,
        DELETE=all_permissions.organizations_change,
    )
    serializer_class = OrganizationMemberUserSerializer

    def get_queryset(self):
        org = generics.get_object_or_404(
            self.request.user.organizations, pk=self.kwargs[self.lookup_field]
        )
        return org.members

    @swagger_auto_schema(tags=["Organizations"])
    def get(self, request, *args, **kwargs):
        return super(OrganizationMemberListAPI, self).get(request, *args, **kwargs)


class OrganizationAPI(
    APIViewVirtualRedirectMixin, APIViewVirtualMethodMixin, generics.RetrieveUpdateAPIView
):
    """
    get:
    Get organization settings

    Retrieve the settings for a specific organization.

    patch:
    Update organization settings

    Update the settings for a specific organization.
    """

    parser_classes = (JSONParser, FormParser, MultiPartParser)
    queryset = Organization.objects.all()
    permission_required = all_permissions.organizations_change
    serializer_class = OrganizationSerializer

    redirect_route = "organizations-dashboard"
    redirect_kwarg = "pk"

    def get_object(self):
        org = generics.get_object_or_404(
            self.request.user.organizations, pk=self.kwargs[self.lookup_field]
        )
        self.check_object_permissions(self.request, org)
        return org

    @swagger_auto_schema(tags=["Organizations"])
    def get(self, request, *args, **kwargs):
        return super(OrganizationAPI, self).get(request, *args, **kwargs)

    @swagger_auto_schema(tags=["Organizations"])
    def patch(self, request, *args, **kwargs):
        return super(OrganizationAPI, self).patch(request, *args, **kwargs)

    @swagger_auto_schema(auto_schema=None)
    def post(self, request, *args, **kwargs):
        return super(OrganizationAPI, self).post(request, *args, **kwargs)

    @swagger_auto_schema(auto_schema=None)
    def put(self, request, *args, **kwargs):
        return super(OrganizationAPI, self).put(request, *args, **kwargs)


class OrganizationInviteAPI(APIView):
    parser_classes = (JSONParser,)
    permission_required = all_permissions.organizations_change

    @swagger_auto_schema(
        tags=["Invites"],
        operation_summary="Get organization invite link",
        responses={200: OrganizationInviteSerializer()},
    )
    def get(self, request, *args, **kwargs):
        org = get_object_with_check_and_log(
            self.request, Organization, pk=request.user.active_organization_id
        )
        self.check_object_permissions(self.request, org)
        invite_url = "{}?token={}".format(reverse("user-signup"), org.token)
        serializer = OrganizationInviteSerializer(
            data={"invite_url": invite_url, "token": org.token}
        )
        serializer.is_valid()
        return Response(serializer.data, status=200)


class OrganizationResetTokenAPI(APIView):
    permission_required = all_permissions.organizations_invite
    parser_classes = (JSONParser,)

    @swagger_auto_schema(
        tags=["Invites"],
        operation_summary="Reset organization token",
        responses={200: OrganizationInviteSerializer()},
    )
    def post(self, request, *args, **kwargs):
        org = request.user.active_organization
        org.reset_token()
        logger.debug(f"New token for organization {org.pk} is {org.token}")
        invite_url = "{}?token={}".format(reverse("user-signup"), org.token)
        serializer = OrganizationInviteSerializer(
            data={"invite_url": invite_url, "token": org.token}
        )
        serializer.is_valid()
        return Response(serializer.data, status=201)



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import logging

from django.db.models import Q
from django.utils import timezone

import drf_yasg.openapi as openapi
from drf_yasg.utils import swagger_auto_schema

from rest_framework.parsers import FormParser, JSONParser, MultiPartParser
from rest_framework import generics
from rest_framework.response import Response
from rest_framework.permissions import IsAuthenticated

from core.utils.common import get_object_with_check_and_log
from core.decorators import permission_required
from core.permissions import all_permissions, ViewClassPermission

from tasks.models import Task, Annotation, Prediction, AnnotationDraft
from core.permissions import get_object_with_permissions, check_object_permissions
from core.mixins import RequestDebugLogMixin
from core.utils.common import bool_from_request
from tasks.serializers import (
    TaskSerializer,
    AnnotationSerializer,
    TaskSimpleSerializer,
    PredictionSerializer,
    TaskWithAnnotationsAndPredictionsAndDraftsSerializer,
    AnnotationDraftSerializer,
)
from projects.models import Project

logger = logging.getLogger(__name__)


class TaskListAPI(generics.ListCreateAPIView):
    """
    post:
    Create task

    Create a new labeling task in Label Studio.
    """

    parser_classes = (JSONParser, FormParser, MultiPartParser)
    queryset = Task.objects.all()
    serializer_class = TaskSerializer

    @swagger_auto_schema(auto_schema=None)
    def get(self, request, *args, **kwargs):
        return super(TaskListAPI, self).get(request, *args, **kwargs)

    def get_serializer_context(self):
        context = super(TaskListAPI, self).get_serializer_context()
        project_id = self.request.data.get("project")
        if project_id:
            context["project"] = generics.get_object_or_404(Project, pk=project_id)
        return context

    @swagger_auto_schema(tags=["Tasks"], request_body=TaskSerializer)
    def post(self, request, *args, **kwargs):
        return super(TaskListAPI, self).post(request, *args, **kwargs)


class TaskAPI(generics.RetrieveUpdateDestroyAPIView):
    """
    get:
    Get task by ID

    Get task data, metadata, annotations and other attributes for a specific labeling task.

    patch:
    Update task

    Update the attributes of an existing labeling task.

    delete:
    Delete task

    Delete a task in Label Studio. This action cannot be undone!
    """

    parser_classes = (JSONParser, FormParser, MultiPartParser)
    queryset = Task.objects.all()

    def get_serializer_class(self):
        # GET => task + annotations + predictions + drafts
        if self.request.method == "GET":
            return TaskWithAnnotationsAndPredictionsAndDraftsSerializer

        # POST, PATCH, PUT
        else:
            return TaskSimpleSerializer

    def retrieve(self, request, *args, **kwargs):
        task = self.get_object()

        # call machine learning api and format response
        if task.project.evaluate_predictions_automatically:
            for ml_backend in task.project.ml_backends.all():
                ml_backend.predict_one_task(task)

        result = self.get_serializer(task).data

        # use proxy inlining to task data (for credential access)
        proxy = bool_from_request(request.GET, "proxy", True)
        result["data"] = task.resolve_uri(result["data"], proxy=proxy)
        return Response(result)

    @swagger_auto_schema(
        tags=["Tasks"],
        manual_parameters=[
            openapi.Parameter(
                name="proxy",
                type=openapi.TYPE_BOOLEAN,
                in_=openapi.IN_QUERY,
                description="Use the proxy parameter inline for credential access to task data",
            )
        ],
    )
    def get(self, request, *args, **kwargs):
        return super(TaskAPI, self).get(request, *args, **kwargs)

    @swagger_auto_schema(tags=["Tasks"], request_body=TaskSimpleSerializer)
    def patch(self, request, *args, **kwargs):
        return super(TaskAPI, self).patch(request, *args, **kwargs)

    @swagger_auto_schema(tags=["Tasks"])
    def delete(self, request, *args, **kwargs):
        return super(TaskAPI, self).delete(request, *args, **kwargs)

    @swagger_auto_schema(auto_schema=None)
    def put(self, request, *args, **kwargs):
        return super(TaskAPI, self).put(request, *args, **kwargs)


class AnnotationAPI(RequestDebugLogMixin, generics.RetrieveUpdateDestroyAPIView):
    """
    get:
    Get annotation by its ID

    Retrieve a specific annotation for a task.

    patch:
    Update annotation

    Update existing attributes on an annotation.

    delete:
    Delete annotation

    Delete an annotation. This action can't be undone!
    """

    parser_classes = (JSONParser, FormParser, MultiPartParser)
    permission_required = ViewClassPermission(
        GET=all_permissions.annotations_view,
        PUT=all_permissions.annotations_change,
        PATCH=all_permissions.annotations_change,
        DELETE=all_permissions.annotations_delete,
    )

    serializer_class = AnnotationSerializer
    queryset = Annotation.objects.all()

    def perform_destroy(self, annotation):
        annotation.delete()

    def update(self, request, *args, **kwargs):
        # save user history with annotator_id, time & annotation result
        annotation_id = self.kwargs["pk"]
        annotation = get_object_with_check_and_log(request, Annotation, pk=annotation_id)

        annotation.task.save()  # refresh task metrics

        return super(AnnotationAPI, self).update(request, *args, **kwargs)

    @swagger_auto_schema(tags=["Annotations"])
    def get(self, request, *args, **kwargs):
        return super(AnnotationAPI, self).get(request, *args, **kwargs)

    @swagger_auto_schema(tags=["Annotations"], auto_schema=None)
    def put(self, request, *args, **kwargs):
        return super(AnnotationAPI, self).put(request, *args, **kwargs)

    @swagger_auto_schema(tags=["Annotations"], request_body=AnnotationSerializer)
    def patch(self, request, *args, **kwargs):
        return super(AnnotationAPI, self).patch(request, *args, **kwargs)

    @swagger_auto_schema(tags=["Annotations"])
    def delete(self, request, *args, **kwargs):
        return super(AnnotationAPI, self).delete(request, *args, **kwargs)


class AnnotationsListAPI(RequestDebugLogMixin, generics.ListCreateAPIView):
    """
    get:
    Get all task annotations

    List all annotations for a task.

    post:
    Create new annotation

    Add annotations to a task like an annotator does.
    """

    parser_classes = (JSONParser, FormParser, MultiPartParser)
    permission_required = ViewClassPermission(
        GET=all_permissions.annotations_view,
        POST=all_permissions.annotations_create,
    )

    serializer_class = AnnotationSerializer

    @swagger_auto_schema(tags=["Annotations"])
    def get(self, request, *args, **kwargs):
        return super(AnnotationsListAPI, self).get(request, *args, **kwargs)

    @swagger_auto_schema(tags=["Annotations"], request_body=AnnotationSerializer)
    def post(self, request, *args, **kwargs):
        return super(AnnotationsListAPI, self).post(request, *args, **kwargs)

    def get_queryset(self):
        task = generics.get_object_or_404(
            Task.objects.for_user(self.request.user), pk=self.kwargs.get("pk", 0)
        )
        return Annotation.objects.filter(Q(task=task) & Q(was_cancelled=False)).order_by("pk")

    def perform_create(self, ser):
        task = get_object_with_check_and_log(self.request, Task, pk=self.kwargs["pk"])
        # annotator has write access only to annotations and it can't be checked it after serializer.save()
        user = self.request.user
        # Release task if it has been taken at work (it should be taken by the same user, or it makes sentry error
        logger.debug(f"User={user} releases task={task}")
        task.release_lock(user)

        # updates history
        update_id = self.request.user.id
        result = ser.validated_data.get("result")
        extra_args = {"task_id": self.kwargs["pk"]}

        # save stats about how well annotator annotations coincide with current prediction
        # only for finished task annotations
        if result is not None:
            prediction = Prediction.objects.filter(
                task=task, model_version=task.project.model_version
            )
            if prediction.exists():
                prediction = prediction.first()
                prediction_ser = PredictionSerializer(prediction).data
            else:
                logger.debug(f"User={self.request.user}: there are no predictions for task={task}")
                prediction_ser = {}
            # serialize annotation
            extra_args.update(
                {
                    "prediction": prediction_ser,
                }
            )

        if "was_cancelled" in self.request.GET:
            extra_args["was_cancelled"] = bool_from_request(
                self.request.GET, "was_cancelled", False
            )

        if "completed_by" not in ser.validated_data:
            extra_args["completed_by"] = self.request.user

        # create annotation
        logger.debug(f"User={self.request.user}: save annotation")
        annotation = ser.save(**extra_args)
        logger.debug(f"Save activity for user={self.request.user}")
        self.request.user.activity_at = timezone.now()
        self.request.user.save()

        # if annotation created from draft - remove this draft
        draft_id = self.request.data.get("draft_id")
        if draft_id is not None:
            logger.debug(f"Remove draft {draft_id} after creating annotation {annotation.id}")
            AnnotationDraft.objects.filter(id=draft_id).delete()

        return annotation


class AnnotationDraftListAPI(RequestDebugLogMixin, generics.ListCreateAPIView):

    parser_classes = (JSONParser, MultiPartParser, FormParser)
    serializer_class = AnnotationDraftSerializer
    permission_required = ViewClassPermission(
        GET=all_permissions.annotations_view,
        POST=all_permissions.annotations_create,
    )
    queryset = AnnotationDraft.objects.all()
    swagger_schema = None

    def filter_queryset(self, queryset):
        task_id = self.kwargs["pk"]
        return queryset.filter(task_id=task_id)

    def perform_create(self, serializer):
        task_id = self.kwargs["pk"]
        annotation_id = self.kwargs.get("annotation_id")
        user = self.request.user
        logger.debug(
            f"User {user} is going to create draft for task={task_id}, annotation={annotation_id}"
        )
        serializer.save(
            task_id=self.kwargs["pk"], annotation_id=annotation_id, user=self.request.user
        )


class AnnotationDraftAPI(RequestDebugLogMixin, generics.RetrieveUpdateDestroyAPIView):

    parser_classes = (JSONParser, MultiPartParser, FormParser)
    serializer_class = AnnotationDraftSerializer
    queryset = AnnotationDraft.objects.all()
    permission_required = ViewClassPermission(
        GET=all_permissions.annotations_view,
        PUT=all_permissions.annotations_change,
        PATCH=all_permissions.annotations_change,
        DELETE=all_permissions.annotations_delete,
    )
    swagger_schema = None



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import logging
import re

logger = logging.getLogger(__name__)

# Put storage prefixes here
uri_regex = r"[\s\'\"]?(?P<uri>(?P<storage>{})://([^/\s]+)/(.*?[^/\s]+/?[^\s\'\">]+))[\s\'\"]?"


def get_uri_via_regex(data, prefixes=("s3", "gs")):
    try:
        uri_regex_prepared = uri_regex.format("|".join(prefixes))
        r_match = re.search(uri_regex_prepared, data)
    except Exception as exc:
        logger.error(f"{data} can't be processed. Reason: {exc}", exc_info=True)
        return None, None
    else:
        if r_match is None:
            logger.warning(
                "{data} does not match uri regex {uri_regex}".format(data=data, uri_regex=uri_regex)
            )
            return None, None
    return r_match.group("uri"), r_match.group("storage")



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
from django.urls import path, include

from . import api
from . import views


app_name = "data_import"

_urlpatterns = []

_api_urlpatterns = [
    path("file-upload/<int:pk>", api.FileUploadAPI.as_view(), name="file-upload-detail")
]

_api_projects_urlpatterns = [
    # import api
    path(
        "<int:pk>/tasks/bulk/", api.TasksBulkCreateAPI.as_view(), name="project-tasks-bulk-upload"
    ),
    path("<int:pk>/import", api.ImportAPI.as_view(), name="project-import"),
    path("<int:pk>/reimport", api.ReImportAPI.as_view(), name="project-reimport"),
    path("<int:pk>/file-uploads", api.FileUploadListAPI.as_view(), name="project-file-upload-list"),
    path("<int:pk>/", api.FileUploadListAPI.as_view(), name="project-file-upload-list"),
]

urlpatterns = [
    path("api/import/", include((_api_urlpatterns, app_name), namespace="api")),
    path("api/projects/", include((_api_projects_urlpatterns, app_name), namespace="api-projects")),
    # special endpoints for serving imported files
    path("data/upload/<str:filename>", api.UploadedFileResponse.as_view(), name="data-upload"),
]



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
# Generated by Django 3.1.4 on 2021-01-21 17:39

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("projects", "__first__"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="FileUpload",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("file", models.FileField(upload_to="upload")),
                (
                    "project",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="file_uploads",
                        to="projects.project",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="file_uploads",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
    ]



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import logging
import ujson as json

from urllib.parse import urlparse
from django.conf import settings
from rest_framework.exceptions import ValidationError
from core.label_config import replace_task_data_undefined_with_config_field


class SkipField(Exception):
    pass


_DATA_TYPES = {
    "Text": [str, int, float],
    "Header": [str, int, float],
    "HyperText": [str],
    "Image": [str, list],
    "Paragraphs": [list],
    "Table": [dict],
    "TimeSeries": [dict, list, str],
    "TimeSeriesChannel": [dict, list, str],
}
logger = logging.getLogger(__name__)


class TaskValidator:
    """Task Validator with project scheme configs validation. It is equal to TaskSerializer from django backend."""

    def __init__(self, project, instance=None):
        self.project = project
        self.instance = instance
        self.annotation_count = 0
        self.prediction_count = 0

    @staticmethod
    def check_data(project, data):
        """Validate data from task['data']"""
        if data is None:
            raise ValidationError("Task is empty (None)")

        replace_task_data_undefined_with_config_field(data, project)

        # iterate over data types from project
        for data_key, data_type in project.data_types.items():
            if data_key not in data:
                raise ValidationError(
                    '"{data_key}" key is expected in task data'.format(data_key=data_key)
                )

            expected_types = _DATA_TYPES.get(data_type, (str,))
            if not isinstance(data[data_key], tuple(expected_types)):
                raise ValidationError(
                    'data["{data_key}"]={data_value} '
                    'is of type "{type}", '
                    'but types "{expected_types}" are expected'.format(
                        data_key=data_key,
                        data_value=data[data_key],
                        type=type(data[data_key]),
                        expected_types=expected_types,
                    )
                )

            if data_type == "List":
                for item in data[data_key]:
                    key = "text"  # FIXME: read key from config (elementValue from List)
                    if key not in item:
                        raise ValidationError("Each item from List must have key " + key)

        return data

    @staticmethod
    def check_data_and_root(project, data, dict_is_root=False):
        """Check data consistent and data is dict with task or dict['task'] is task

        :param project:
        :param data:
        :param dict_is_root:
        :return:
        """
        try:
            TaskValidator.check_data(project, data)
        except ValidationError as e:
            if dict_is_root:
                raise ValidationError(
                    e.detail[0] + " [assume: item as is = task root with values] "
                )
            else:
                raise ValidationError(
                    e.detail[0] + ' [assume: item["data"] = task root with values]'
                )

    @staticmethod
    def check_allowed(task):
        # task is required
        if "data" not in task:
            return False

        # everything is ok
        return True

    @staticmethod
    def raise_if_wrong_class(task, key, class_def):
        if key in task and not isinstance(task[key], class_def):
            raise ValidationError(
                "Task[{key}] must be {class_def}".format(key=key, class_def=class_def)
            )

    def validate(self, task):
        """Validate whole task with task['data'] and task['annotations']. task['predictions']"""
        # task is class
        if hasattr(task, "data"):
            self.check_data_and_root(self.project, task.data)
            return task

        # self.instance is loaded by get_object of view
        if self.instance and hasattr(self.instance, "data"):
            if isinstance(self.instance.data, dict):
                data = self.instance.data
            elif isinstance(self.instance.data, str):
                try:
                    data = json.loads(self.instance.data)
                except ValueError as e:
                    raise ValidationError("Can't parse task data: " + str(e))
            else:
                raise ValidationError(
                    'Field "data" must be string or dict, but not "'
                    + type(self.instance.data)
                    + '"'
                )
            self.check_data_and_root(self.instance.project, data)
            return task

        # check task is dict
        if not isinstance(task, dict):
            raise ValidationError(
                'Task root must be dict with "data", "meta", "annotations", "predictions" fields'
            )

        # task[data] | task[annotations] | task[predictions] | task[meta]
        if self.check_allowed(task):
            # task[data]
            self.raise_if_wrong_class(task, "data", (dict, list))
            self.check_data_and_root(self.project, task["data"])

            # task[annotations]: we can't use AnnotationSerializer for validation
            # because it's much different with validation we need here
            self.raise_if_wrong_class(task, "annotations", list)
            for annotation in task.get("annotations", []):
                ok = "result" in annotation
                if not ok:
                    raise ValidationError('Annotation must have "result" fields')

                # check result is list
                if not isinstance(annotation.get("result", []), list):
                    raise ValidationError('"result" field in annotation must be list')

            # task[predictions]
            self.raise_if_wrong_class(task, "predictions", list)
            for prediction in task.get("predictions", []):
                ok = "result" in prediction
                if not ok:
                    raise ValidationError('Prediction must have "result" fields')

                # check result is list
                if not isinstance(prediction.get("result", []), list):
                    raise ValidationError('"result" field in prediction must be list')

            # task[meta]
            self.raise_if_wrong_class(task, "meta", (dict, list))

        # task is data as is, validate task as data and move it to task['data']
        else:
            self.check_data_and_root(self.project, task, dict_is_root=True)
            task = {"data": task}

        return task

    @staticmethod
    def format_error(i, detail, item):
        if len(detail) == 1:
            code = (str(detail[0].code + " ")) if detail[0].code != "invalid" else ""
            return "Error {code} at item {i}: {detail} :: {item}".format(
                code=code, i=i, detail=detail[0], item=item
            )
        else:
            errors = ", ".join(detail)
            codes = str([d.code for d in detail])
            return "Errors {codes} at item {i}: {errors} :: {item}".format(
                codes=codes, i=i, errors=errors, item=item
            )

    def to_internal_value(self, data):
        """Body of run_validation for all data items"""
        if data is None:
            raise ValidationError("All tasks are empty (None)")

        if not isinstance(data, list):
            raise ValidationError("data is not a list")

        if len(data) == 0:
            raise ValidationError("data is empty")

        ret, errors = [], []
        self.annotation_count, self.prediction_count = 0, 0
        for i, item in enumerate(data):
            try:
                validated = self.validate(item)
            except ValidationError as exc:
                error = self.format_error(i, exc.detail, item)
                errors.append(error)
                # do not print to user too many errors
                if len(errors) >= 100:
                    errors[99] = "..."
                    break
            else:
                ret.append(validated)
                errors.append({})

                if "annotations" in item:
                    self.annotation_count += len(item["annotations"])
                if "predictions" in item:
                    self.prediction_count += len(item["predictions"])

        if any(errors):
            logger.warning("Can't deserialize tasks due to " + str(errors))
            raise ValidationError(errors)

        return ret


def is_url(string):
    try:
        result = urlparse(string.strip())
        return all([result.scheme, result.netloc])
    except ValueError:
        return False



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import logging
import django_rq
import json

from django.utils import timezone
from django.db import models, transaction
from django.utils.translation import gettext_lazy as _
from django_rq import job

from tasks.models import Task
from tasks.serializers import PredictionSerializer, AnnotationSerializer
from data_export.serializers import ExportDataSerializer


from core.redis import redis_connected
from core.utils.common import get_bool_env


logger = logging.getLogger(__name__)


class Storage(models.Model):
    title = models.CharField(_("title"), null=True, max_length=256, help_text="Cloud storage title")
    description = models.TextField(
        _("description"), null=True, blank=True, help_text="Cloud storage description"
    )
    project = models.ForeignKey(
        "projects.Project", related_name="%(app_label)s_%(class)ss", on_delete=models.CASCADE
    )
    created_at = models.DateTimeField(_("created at"), auto_now_add=True, help_text="Creation time")
    last_sync = models.DateTimeField(
        _("last sync"), null=True, blank=True, help_text="Last sync finished time"
    )
    last_sync_count = models.PositiveIntegerField(
        _("last sync count"), null=True, blank=True, help_text="Count of tasks synced last time"
    )

    def validate_connection(self, client=None):
        pass

    def has_permission(self, user):
        if self.project.has_permission(user):
            return True
        return False

    class Meta:
        abstract = True


class ImportStorage(Storage):
    def iterkeys(self):
        return iter(())

    def get_data(self, key):
        raise NotImplementedError

    def resolve_uri(self, data):
        return

    def resolve_task_data_uri(self, task_data):
        out = {}
        for key, data in task_data.items():
            if not isinstance(data, str):
                out[key] = data
            resolved_uri = self.resolve_uri(data)
            if resolved_uri:
                out[key] = resolved_uri
            else:
                out[key] = data
        return out

    def _scan_and_create_links(self, link_class):
        tasks_created = 0
        for key in self.iterkeys():
            logger.debug(f"Scanning key {key}")

            # skip if task already exists
            if link_class.exists(key, self):
                logger.debug(f"{self.__class__.__name__} link {key} already exists")
                continue

            logger.debug(f"{self}: found new key {key}")
            try:
                data = self.get_data(key)
            except (UnicodeDecodeError, json.decoder.JSONDecodeError) as exc:
                logger.error(exc, exc_info=True)
                raise ValueError(
                    f'Error loading JSON from file "{key}".\nIf you\'re trying to import non-JSON data '
                    f"(images, audio, text, etc.), edit storage settings and enable "
                    f'"Treat every bucket object as a source file"'
                )

            # predictions
            predictions = data.get("predictions", [])
            if predictions:
                if "data" not in data:
                    raise ValueError(
                        'If you use "predictions" field in the task, '
                        'you must put "data" field in the task too'
                    )

            # annotations
            annotations = data.get("annotations", [])
            if annotations:
                if "data" not in data:
                    raise ValueError(
                        'If you use "annotations" field in the task, '
                        'you must put "data" field in the task too'
                    )

            if "data" in data and isinstance(data["data"], dict):
                data = data["data"]

            with transaction.atomic():
                task = Task.objects.create(data=data, project=self.project)
                link_class.create(task, key, self)
                logger.debug(
                    f"Create {self.__class__.__name__} link with key={key} for task={task}"
                )
                tasks_created += 1

                # add predictions
                logger.debug(f"Create {len(predictions)} predictions for task={task}")
                for prediction in predictions:
                    prediction["task"] = task.id
                prediction_ser = PredictionSerializer(data=predictions, many=True)
                prediction_ser.is_valid(raise_exception=True)
                prediction_ser.save()

                # add annotations
                logger.debug(f"Create {len(annotations)} annotations for task={task}")
                for annotation in annotations:
                    annotation["task"] = task.id
                annotation_ser = AnnotationSerializer(data=annotations, many=True)
                annotation_ser.is_valid(raise_exception=True)
                annotation_ser.save()

        self.last_sync = timezone.now()
        self.last_sync_count = tasks_created
        self.save()

    def scan_and_create_links(self):
        """This is proto method - you can override it, or just replace ImportStorageLink by your own model"""
        self._scan_and_create_links(ImportStorageLink)

    def sync(self):
        if redis_connected():
            queue = django_rq.get_queue("default")
            job = queue.enqueue(sync_background, self.__class__, self.id)
            # job_id = sync_background.delay()  # TODO: @niklub: check this fix
            logger.info(f"Storage sync background job {job.id} for storage {self} has been started")
        else:
            logger.info(f"Start syncing storage {self}")
            self.scan_and_create_links()

    class Meta:
        abstract = True


@job("default")
def sync_background(storage_class, storage_id):
    storage = storage_class.objects.get(id=storage_id)
    storage.scan_and_create_links()


class ExportStorage(Storage):
    def _get_serialized_data(self, annotation):
        if get_bool_env("FUTURE_SAVE_TASK_TO_STORAGE", default=False):
            # export task with annotations
            return ExportDataSerializer(annotation.task).data
        else:
            from io_storages.serializers import StorageAnnotationSerializer

            # deprecated functionality - save only annotation
            return StorageAnnotationSerializer(annotation).data

    def save_annotation(self, annotation):
        raise NotImplementedError

    class Meta:
        abstract = True


class ImportStorageLink(models.Model):

    task = models.OneToOneField(
        "tasks.Task", on_delete=models.CASCADE, related_name="%(app_label)s_%(class)s"
    )
    key = models.TextField(_("key"), null=False, help_text="External link key")
    object_exists = models.BooleanField(
        _("object exists"),
        help_text="Whether object under external link still exists",
        default=True,
    )
    created_at = models.DateTimeField(_("created at"), auto_now_add=True, help_text="Creation time")

    @classmethod
    def exists(cls, key, storage):
        return cls.objects.filter(key=key, storage=storage.id).exists()

    @classmethod
    def create(cls, task, key, storage):
        link, created = cls.objects.get_or_create(
            task_id=task.id, key=key, storage=storage, object_exists=True
        )
        return link

    def has_permission(self, user):
        if self.task.has_permission(user):
            return True
        return False

    class Meta:
        abstract = True


class ExportStorageLink(models.Model):

    annotation = models.OneToOneField(
        "tasks.Annotation", on_delete=models.CASCADE, related_name="%(app_label)s_%(class)s"
    )
    object_exists = models.BooleanField(
        _("object exists"),
        help_text="Whether object under external link still exists",
        default=True,
    )
    created_at = models.DateTimeField(_("created at"), auto_now_add=True, help_text="Creation time")

    @property
    def key(self):
        if get_bool_env("FUTURE_SAVE_TASK_TO_STORAGE", default=False):
            return str(self.annotation.task.id)
        return str(self.annotation.id)

    @classmethod
    def exists(cls, annotation, storage):
        return cls.objects.filter(annotation=annotation.id, storage=storage.id).exists()

    @classmethod
    def create(cls, annotation, storage):
        link, created = cls.objects.get_or_create(
            annotation=annotation, storage=storage, object_exists=True
        )
        return link

    def has_permission(self, user):
        if self.annotation.has_permission(user):
            return True
        return False

    class Meta:
        abstract = True



# Generated by Django 3.1.4 on 2021-03-04 14:23
import django

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    atomic = False

    dependencies = [
        ("io_storages", "0001_squashed_0002_auto_20210302_1827"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("tasks", "0001_squashed_0041_taskcompletionhistory_was_cancelled"),
    ]

    operations = [
        migrations.RenameModel(
            old_name="TaskCompletion",
            new_name="Annotation",
        ),
        migrations.RenameModel(
            old_name="TaskCompletionDraft",
            new_name="AnnotationDraft",
        ),
        migrations.RenameField(
            model_name="review",
            old_name="completion",
            new_name="annotation",
        ),
        migrations.RenameField(
            model_name="annotationdraft",
            old_name="completion",
            new_name="annotation",
        ),
        migrations.AlterField(
            model_name="annotationdraft",
            name="annotation",
            field=models.ForeignKey(
                blank=True,
                help_text="Corresponding annotation for this draft",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="drafts",
                to="tasks.annotation",
            ),
        ),
        migrations.AlterField(
            model_name="task",
            name="file_upload",
            field=models.ForeignKey(
                blank=True,
                help_text="Uploaded file used as data source for this task",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="tasks",
                to="data_import.fileupload",
            ),
        ),
        migrations.AlterField(
            model_name="task",
            name="is_labeled",
            field=models.BooleanField(
                default=False,
                help_text="True if the annotation number for this task is greater or equal to the project maximum_annotations",
                verbose_name="is_labeled",
            ),
        ),
        migrations.AlterField(
            model_name="annotation",
            name="completed_by",
            field=models.ForeignKey(
                help_text="User ID who made this annotation",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="annotations",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AlterField(
            model_name="annotation",
            name="honeypot",
            field=models.BooleanField(
                default=False,
                help_text="This annotation is Ground Truth (honeypot)",
                verbose_name="honeypot",
            ),
        ),
        migrations.AlterField(
            model_name="annotation",
            name="lead_time",
            field=models.FloatField(
                default=None,
                help_text="How much time spent to solve the annotation",
                null=True,
                verbose_name="lead time",
            ),
        ),
        migrations.AlterField(
            model_name="annotation",
            name="prediction",
            field=models.JSONField(
                default=dict,
                help_text="Prediction viewed at the time of annotation",
                null=True,
                verbose_name="prediction",
            ),
        ),
        migrations.AlterField(
            model_name="annotation",
            name="prediction_equal_score",
            field=models.FloatField(
                default=0.0,
                help_text="Comparison result with prediction viewed at the time of annotation",
                verbose_name="prediction_equal_score",
            ),
        ),
        migrations.AlterField(
            model_name="annotation",
            name="result_count",
            field=models.IntegerField(
                default=0,
                help_text="Results inside of annotation counter",
                verbose_name="result count",
            ),
        ),
        migrations.AlterField(
            model_name="annotation",
            name="task",
            field=models.ForeignKey(
                help_text="Corresponding task for this annotation",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="annotations",
                to="tasks.task",
            ),
        ),
        migrations.AlterField(
            model_name="annotationdraft",
            name="lead_time",
            field=models.FloatField(
                help_text="how much time spent to solve the annotation", verbose_name="lead time"
            ),
        ),
    ]



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
"""
Django Base settings for Label Studio.

For more information on this file, see
https://docs.djangoproject.com/en/3.1/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/3.1/ref/settings/
"""
import os
import re

from core.utils.io import get_data_dir
from core.utils.params import get_bool_env, get_env

# Hostname is used for proper path generation to the resources, pages, etc
HOSTNAME = get_env("HOST", "")
if HOSTNAME:
    if not HOSTNAME.startswith("http://") and not HOSTNAME.startswith("https://"):
        print(
            "! HOST variable found in environment, but it must start with http:// or https://, ignore it:",
            HOSTNAME,
        )
        HOSTNAME = ""
    else:
        print("=> Hostname correctly is set to:", HOSTNAME)
        if HOSTNAME.endswith("/"):
            HOSTNAME = HOSTNAME[0:-1]

        # for django url resolver
        if HOSTNAME:
            # http[s]://domain.com:8080/script_name => /script_name
            pattern = re.compile(r"^http[s]?:\/\/([^:\/\s]+(:\d*)?)(.*)?")
            match = pattern.match(HOSTNAME)
            FORCE_SCRIPT_NAME = match.group(3)
            if FORCE_SCRIPT_NAME:
                print("=> Django URL prefix is set to:", FORCE_SCRIPT_NAME)

INTERNAL_PORT = "8080"

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "$(fefwefwef13;LFK{P!)@#*!)kdsjfWF2l+i5e3t(8a1n"

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = get_bool_env("DEBUG", True)

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

# Base path for media root and other uploaded files
BASE_DATA_DIR = get_env("BASE_DATA_DIR", get_data_dir())
os.makedirs(BASE_DATA_DIR, exist_ok=True)
print("=> Database and media directory:", BASE_DATA_DIR)

# Databases
# https://docs.djangoproject.com/en/2.1/ref/settings/#databases
DJANGO_DB_SQLITE = "sqlite"
DJANGO_DB = "default"
DATABASE_NAME_DEFAULT = os.path.join(BASE_DATA_DIR, "label_studio_withoutsignin.sqlite3")
DATABASE_NAME = get_env("DATABASE_NAME", DATABASE_NAME_DEFAULT)
DATABASES_ALL = {
    "default": {
        "ENGINE": "django.db.backends.postgresql",
        "USER": get_env("POSTGRE_USER", "postgres"),
        "PASSWORD": get_env("POSTGRE_PASSWORD", "postgres"),
        "NAME": get_env("POSTGRE_NAME", "postgres"),
        "HOST": get_env("POSTGRE_HOST", "localhost"),
        "PORT": int(get_env("POSTGRE_PORT", "5432")),
    },
    DJANGO_DB_SQLITE: {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": DATABASE_NAME,
        "OPTIONS": {
            # 'timeout': 20,
        },
    },
}
DATABASES = {"default": DATABASES_ALL.get(get_env("DJANGO_DB", "default"))}

LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "standard": {
            "format": "[%(asctime)s] [%(name)s::%(funcName)s::%(lineno)d] [%(levelname)s] %(message)s",
        },
        "message_only": {
            "format": "%(message)s",
        },
        "rq_console": {
            "format": "%(asctime)s %(message)s",
            "datefmt": "%H:%M:%S",
        },
    },
    "handlers": {
        "console_raw": {
            "level": get_env("LOG_LEVEL", "DEBUG"),
            "class": "logging.StreamHandler",
        },
        "console": {
            "level": get_env("LOG_LEVEL", "DEBUG"),
            "class": "logging.StreamHandler",
            "formatter": "standard",
        },
        "rq_console": {
            "level": "DEBUG",
            "class": "rq.utils.ColorizingStreamHandler",
            "formatter": "rq_console",
            "exclude": ["%(asctime)s"],
        },
    },
    "root": {
        "handlers": ["console"],
        "level": get_env("LOG_LEVEL", "DEBUG"),
    },
}

INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "django.contrib.humanize",
    "drf_yasg",
    "corsheaders",
    "django_extensions",
    "django_rq",
    "django_filters",
    "rules",
    "annoying",
    "rest_framework",
    "rest_framework_swagger",
    "rest_framework.authtoken",
    "drf_generators",
    "core",
    "users",
    "organizations",
    "data_import",
    "projects",
    "tasks",
    "data_manager",
    "io_storages",
    "ml",
]

MIDDLEWARE = [
    "corsheaders.middleware.CorsMiddleware",
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.locale.LocaleMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "core.middleware.DisableCSRF",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    "core.middleware.CommonMiddlewareAppendSlashWithoutRedirect",  # instead of 'CommonMiddleware'
    "core.middleware.CommonMiddleware",
    "core.middleware.DRFResponseFormatter",
    "django_user_agents.middleware.UserAgentMiddleware",
    "core.middleware.SetSessionUIDMiddleware",
    "core.middleware.ContextLogMiddleware",
    "core.middleware.DatabaseIsLockedRetryMiddleware",
]

REST_FRAMEWORK = {
    "DEFAULT_FILTER_BACKENDS": ["django_filters.rest_framework.DjangoFilterBackend"],
    "DEFAULT_AUTHENTICATION_CLASSES": (
        "rest_framework.authentication.TokenAuthentication",
        "rest_framework.authentication.SessionAuthentication",
    ),
    "DEFAULT_PERMISSION_CLASSES": [
        "core.api_permissions.HasObjectPermission",
        "rest_framework.permissions.IsAuthenticated",
    ],
    "EXCEPTION_HANDLER": "core.utils.common.custom_exception_handler",
    "DEFAULT_RENDERER_CLASSES": ("rest_framework.renderers.JSONRenderer",),
    "DEFAULT_VERSIONING_CLASS": "rest_framework.versioning.NamespaceVersioning",
}

# CORS & Host settings
INTERNAL_IPS = [  # django debug toolbar for django==2.2 requirement
    "127.0.0.1",
    "localhost",
]
CORS_ORIGIN_ALLOW_ALL = True
CORS_ALLOW_METHODS = [
    "DELETE",
    "GET",
    "OPTIONS",
    "PATCH",
    "POST",
    "PUT",
]
ALLOWED_HOSTS = ["*"]

# Auth modules
AUTH_USER_MODEL = "users.User"
AUTHENTICATION_BACKENDS = [
    "rules.permissions.ObjectPermissionBackend",
    "django.contrib.auth.backends.ModelBackend",
]
USE_USERNAME_FOR_LOGIN = False

DISABLE_SIGNUP_WITHOUT_LINK = get_bool_env("DISABLE_SIGNUP_WITHOUT_LINK", False)

# Password validation:
# https://docs.djangoproject.com/en/2.1/ref/settings/#auth-password-validators
AUTH_PASSWORD_VALIDATORS = [
    {"NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator"},
    {"NAME": "django.contrib.auth.password_validation.MinimumLengthValidator"},
    {"NAME": "django.contrib.auth.password_validation.CommonPasswordValidator"},
    {"NAME": "django.contrib.auth.password_validation.NumericPasswordValidator"},
]

# Django templates
TEMPLATES_DIR = os.path.join(os.path.dirname(BASE_DIR), "templates")  # ../../from_this = 'web' dir
TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [TEMPLATES_DIR],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
                "core.context_processors.settings",
            ],
            "builtins": ["django.templatetags.i18n"],
        },
    }
]

# RQ
RQ_QUEUES = {"default": {"HOST": "localhost", "PORT": 6379, "DB": 0, "DEFAULT_TIMEOUT": 180}}

# Swagger: automatic API documentation
SWAGGER_SETTINGS = {
    "SECURITY_DEFINITIONS": {"token": {"type": "token", "url": "/business/account/#/token"}},
    "APIS_SORTER": "alpha",
    "SUPPORTED_SUBMIT_METHODS": ["get", "post", "put", "delete", "patch"],
    # "DEFAULT_AUTO_SCHEMA_CLASS": "core.utils.CustomAutoSchema",
    "OPERATIONS_SORTER": "alpha",
}

SENTRY_FE = None

ROOT_URLCONF = "core.urls"
WSGI_APPLICATION = "core.wsgi.application"
GRAPHIQL = True

# Internationalization
# https://docs.djangoproject.com/en/2.1/topics/i18n/
LANGUAGE_CODE = "en-us"
TIME_ZONE = "UTC"
USE_I18N = False
USE_L10N = True
USE_TZ = True

# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/2.1/howto/static-files/
STATIC_URL = "/static/"
# if FORCE_SCRIPT_NAME:
#    STATIC_URL = FORCE_SCRIPT_NAME + STATIC_URL
print(f"=> Static URL is set to: {STATIC_URL}")

STATIC_ROOT = os.path.join(BASE_DIR, "static_build")
STATICFILES_DIRS = [os.path.join(BASE_DIR, "static")]
STATICFILES_FINDERS = (
    "django.contrib.staticfiles.finders.FileSystemFinder",
    "django.contrib.staticfiles.finders.AppDirectoriesFinder",
)
STATICFILES_STORAGE = "core.storage.SkipMissedManifestStaticFilesStorage"

# Sessions and CSRF
SESSION_COOKIE_SECURE = bool(int(get_env("SESSION_COOKIE_SECURE", True)))
CSRF_COOKIE_SECURE = bool(int(get_env("CSRF_COOKIE_SECURE", SESSION_COOKIE_SECURE)))
CSRF_COOKIE_HTTPONLY = bool(int(get_env("CSRF_COOKIE_HTTPONLY", SESSION_COOKIE_SECURE)))

# user media files
MEDIA_ROOT = os.path.join(BASE_DATA_DIR, "media")
MEDIA_URL = "/data/"
UPLOAD_DIR = "upload"
AVATAR_PATH = "avatars"

# project exports
EXPORT_DIR = os.path.join(BASE_DATA_DIR, "export")
EXPORT_URL_ROOT = "/export/"
os.makedirs(EXPORT_DIR, exist_ok=True)

# file / task size limits
DATA_UPLOAD_MAX_MEMORY_SIZE = int(get_env("DATA_UPLOAD_MAX_MEMORY_SIZE", 50 * 1024 * 1024))
TASKS_MAX_NUMBER = 250000
TASKS_MAX_FILE_SIZE = DATA_UPLOAD_MAX_MEMORY_SIZE

TASK_LOCK_TTL = int(get_env("TASK_LOCK_TTL")) if get_env("TASK_LOCK_TTL") else None
TASK_LOCK_DEFAULT_TTL = int(get_env("TASK_LOCK_DEFAULT_TTL", 3600))

# Email backend
FROM_EMAIL = "Label Studio <hello@labelstud.io>"
EMAIL_BACKEND = "django.core.mail.backends.dummy.EmailBackend"

LOCAL_FILES_SERVING_ENABLED = get_bool_env("LOCAL_FILES_SERVING_ENABLED", default=False)

""" React Libraries: do not forget to change this dir in /etc/nginx/nginx.conf """
# EDITOR = label-studio-frontend repository
EDITOR_ROOT = os.path.join(BASE_DIR, "../frontend/dist/lsf")
# DM = data manager (included into FRONTEND due npm building, we need only version.json file from there)
DM_ROOT = os.path.join(BASE_DIR, "../frontend/dist/dm")
# FRONTEND = GUI for django backend
REACT_APP_ROOT = os.path.join(BASE_DIR, "../frontend/dist/react-app")

# per project settings
BATCH_SIZE = 1000
PROJECT_TITLE_MIN_LEN = 3
PROJECT_TITLE_MAX_LEN = 50
LOGIN_REDIRECT_URL = "/"
LOGIN_URL = "/"
MIN_GROUND_TRUTH = 10
DATA_UNDEFINED_NAME = "$undefined$"
LICENSE = {}
VERSIONS = {}
VERSION_EDITION = "Community Edition"

CREATE_ORGANIZATION = "organizations.functions.create_organization"
GET_OBJECT_WITH_CHECK_AND_LOG = "core.utils.get_object.get_object_with_check_and_log"
SAVE_USER = "users.functions.save_user"
USER_SERIALIZER = "users.serializers.BaseUserSerializer"
DATA_MANAGER_ANNOTATIONS_MAP = {}
DATA_MANAGER_ACTIONS = {}
USER_LOGIN_FORM = "users.forms.LoginForm"
PROJECT_MIXIN = "core.mixins.DummyModelMixin"
TASK_MIXIN = "core.mixins.DummyModelMixin"
ANNOTATION_MIXIN = "core.mixins.DummyModelMixin"
ORGANIZATION_MIXIN = "core.mixins.DummyModelMixin"
USER_MIXIN = "users.mixins.UserMixin"


def project_delete(project):
    project.delete()


def user_auth(user_model, email, password):
    return None


def collect_versions(**kwargs):
    return {}


PROJECT_DELETE = project_delete
USER_AUTH = user_auth
COLLECT_VERSIONS = collect_versions

# fix a problem with Windows mimetypes for JS and PNG
import mimetypes

mimetypes.add_type("application/javascript", ".js", True)
mimetypes.add_type("image/png", ".png", True)

print()  # just empty line



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import os
import logging

from datetime import datetime
from django import forms
from django.contrib import auth
from django.conf import settings

from users.models import User


PASS_MAX_LENGTH = 64
PASS_MIN_LENGTH = 8
USERNAME_MAX_LENGTH = 30
DISPLAY_NAME_LENGTH = 100
USERNAME_LENGTH_ERR = "Please enter a username 30 characters or fewer in length"
DISPLAY_NAME_LENGTH_ERR = "Please enter a display name 100 characters or fewer in length"
PASS_LENGTH_ERR = "Please enter a password 8-12 characters in length"
INVALID_USER_ERROR = "The email and password you entered don't match."

logger = logging.getLogger(__name__)


class LoginForm(forms.Form):
    """For logging in to the app and all - session based"""

    # use username instead of email when LDAP enabled
    email = (
        forms.CharField(label="User")
        if settings.USE_USERNAME_FOR_LOGIN
        else forms.EmailField(label="Email")
    )
    password = forms.CharField(widget=forms.PasswordInput())

    def clean(self, *args, **kwargs):
        cleaned = super(LoginForm, self).clean()
        email = cleaned.get("email", "").lower()
        password = cleaned.get("password", "")

        # advanced way for user auth
        user = settings.USER_AUTH(User, email, password)

        # regular access
        if user is None:
            user = auth.authenticate(email=email, password=password)

        if user and user.is_active:
            return {"user": user}
        else:
            raise forms.ValidationError(INVALID_USER_ERROR)


class ForgotPasswordForm(forms.Form):
    email = forms.EmailField()

    def clean_email(self):
        email = self.cleaned_data["email"].lower()
        try:
            User.objects.get(email=email)
        except User.DoesNotExist:
            raise forms.ValidationError("A user matching that email address was not found")
        return email


class BasePasswordForm(forms.Form):
    password = forms.CharField(
        max_length=PASS_MAX_LENGTH, error_messages={"required": PASS_LENGTH_ERR}
    )

    def clean_password(self):
        password = self.cleaned_data["password"]
        if len(password) < PASS_MIN_LENGTH:
            raise forms.ValidationError(PASS_LENGTH_ERR)
        return password


class ResetPasswordForm(BasePasswordForm):
    pass


class UserSignupForm(forms.Form):
    email = forms.EmailField(label="Work Email", error_messages={"required": "Invalid email"})
    password = forms.CharField(
        max_length=PASS_MAX_LENGTH,
        error_messages={"required": PASS_LENGTH_ERR},
        widget=forms.TextInput(attrs={"type": "password"}),
    )

    def clean_password(self):
        password = self.cleaned_data["password"]
        if len(password) < PASS_MIN_LENGTH:
            raise forms.ValidationError(PASS_LENGTH_ERR)
        return password

    def clean_username(self):
        username = self.cleaned_data.get("username")
        if username and User.objects.filter(username=username.lower()).exists():
            raise forms.ValidationError("User with username already exists")
        return username

    def clean_email(self):
        email = self.cleaned_data.get("email").lower()

        if email and User.objects.filter(email=email).exists():
            raise forms.ValidationError("User with email already exists")

        return email

    def save(self):
        cleaned = self.cleaned_data
        password = cleaned["password"]
        email = cleaned["email"].lower()
        user = User.objects.create_user(email, password)
        return user


class UserProfileForm(forms.ModelForm):
    """This form is used in profile account pages"""

    class Meta:
        model = User
        fields = ("first_name", "last_name", "phone")
